import unittest

import fqr

from fqr . loggers import lib

from .. import mocking

from . import cfg


class Constants(cfg.Constants):
    """Constant values specific to this file."""


class TestLogger(unittest.TestCase):
    """Fixture for testing logger."""

    def setUp(self) -> None:
        self.log = fqr.log
        self.msg_str = 'example'
        self.msg_dict = {'str': 'example', 'a': 2}
        self.msg_cls = mocking.examples.Pet
        self.msg_object_= mocking.examples.Pet(
            id_='abc1234',
            name='Fido',
            type='dog',
            )
        return super().setUp()

    def test_01_log(self):
        """Test str logging."""

        msg = self.msg_str
        level = lib.logging.DEBUG
        expected_output = lib.textwrap.indent(
            lib.json.dumps(
                fqr.loggers.utl.parse_incoming_log_message(msg, level),
                default=fqr.core.strings.utl.convert_for_repr,
                indent=Constants.INDENT,
                sort_keys=True
                ),
            Constants.INDENT * ' '
            )
        with self.assertLogs(self.log, level) as logger:
            fqr.log.debug(msg)
            self.assertEqual(logger.records[0].msg, expected_output)

    def test_02_log(self):
        """Test dict logging."""

        msg = self.msg_dict
        level = lib.logging.INFO
        expected_output = lib.textwrap.indent(
            lib.json.dumps(
                fqr.loggers.utl.parse_incoming_log_message(msg, level),
                default=fqr.core.strings.utl.convert_for_repr,
                indent=Constants.INDENT,
                sort_keys=True
                ),
            Constants.INDENT * ' '
            )
        with self.assertLogs(self.log, level) as logger:
            fqr.log.info(msg)
            self.assertEqual(logger.records[0].msg, expected_output)

    def test_03_log(self):
        """Test cls logging."""

        msg = self.msg_cls
        level = lib.logging.WARNING
        expected_output = lib.textwrap.indent(
            lib.json.dumps(
                fqr.loggers.utl.parse_incoming_log_message(msg, level),
                default=fqr.core.strings.utl.convert_for_repr,
                indent=Constants.INDENT,
                sort_keys=True
                ),
            Constants.INDENT * ' '
            )
        with self.assertLogs(self.log, level) as logger:
            fqr.log.warning(msg)
            self.assertEqual(logger.records[0].msg, expected_output)

    def test_04_log(self):
        """Test object logging."""

        msg = self.msg_object_
        level = lib.logging.ERROR
        expected_output = lib.textwrap.indent(
            lib.json.dumps(
                fqr.loggers.utl.parse_incoming_log_message(msg, level),
                default=fqr.core.strings.utl.convert_for_repr,
                indent=Constants.INDENT,
                sort_keys=True
                ),
            Constants.INDENT * ' '
            )
        with self.assertLogs(self.log, level) as logger:
            fqr.log.error(msg)
            self.assertEqual(logger.records[0].msg, expected_output)

    def test_05_log(self):
        """Test exc logging."""

        msg = self.msg_str
        level = lib.logging.ERROR
        parsed = fqr.loggers.utl.parse_incoming_log_message(msg, level)
        with self.assertLogs(self.log, level) as logger:
            try:
                _ = 1 / 0
            except Exception:
                fqr.log.error(msg)
                expected_output = lib.textwrap.indent(
                    lib.json.dumps(
                        parsed,
                        default=fqr.core.strings.utl.convert_for_repr,
                        indent=Constants.INDENT,
                        sort_keys=True
                        ),
                    Constants.INDENT * ' '
                    )
            self.assertEqual(logger.records[0].msg, expected_output)

    def test_06_log(self):
        """Test exc logging."""

        msg = self.msg_str
        level = lib.logging.DEBUG
        parsed = fqr.loggers.utl.parse_incoming_log_message(msg, level)
        with self.assertLogs(self.log, level) as logger:
            try:
                _ = 1 / 0
            except Exception as e:
                self.log.debug(msg, exc_info=e)
                expected_output = lib.textwrap.indent(
                    lib.json.dumps(
                        parsed,
                        default=fqr.core.strings.utl.convert_for_repr,
                        indent=Constants.INDENT,
                        sort_keys=True
                        ),
                    Constants.INDENT * ' '
                    )
            self.assertEqual(logger.records[0].msg, expected_output)

    def test_07_print(self):
        """Test first print + warning interception."""

        msg = self.msg_str
        level = lib.logging.WARNING
        parsed = fqr.loggers.utl.parse_incoming_log_message(
            Constants.WARN_MSG,
            level
            )
        with self.assertLogs(self.log, level) as logger:
            print(msg)
            expected_output = lib.textwrap.indent(
                lib.json.dumps(
                    parsed,
                    default=fqr.core.strings.utl.convert_for_repr,
                    indent=Constants.INDENT,
                    sort_keys=True
                    ),
                Constants.INDENT * ' '
                )
            self.assertEqual(
                logger.records[0].msg,
                expected_output
                )

    def test_08_print(self):
        """Test subsequent print interceptions."""

        msg = self.msg_str
        with self.assertWarns(Warning) as emitter:
            with self.assertNoLogs(self.log):
                print(msg)
                self.assertEqual(
                    emitter.warnings[0].message.args[0],
                    Constants.WARN_MSG
                    )

    def test_09_parse_log_record(self):
        """Test idempotent `LogRecord` parsing."""

        msg = fqr.loggers.typ.LogRecord(message='test')
        self.assertEqual(
            msg,
            fqr.loggers.utl.parse_incoming_log_message(msg, 20)
            )

    def test_10_parse_log_record(self):
        """Test array msg for `LogRecord` parsing."""

        msg = ('test', )
        self.assertEqual(
            fqr.loggers.typ.LogRecord(message=msg),
            fqr.loggers.utl.parse_incoming_log_message(msg, 20)
            )


class TestDeployedLogger(unittest.TestCase):
    """Fixture for testing logger in higher environments."""

    def setUp(self) -> None:
        Constants.ENV = 'dev'
        _print = __builtins__['print']

        def _reprint(*args: lib.t.Any, **kwargs: lib.t.Any) -> None:
            if Constants.ENV in Constants.DEPLOY_ENVS:
                lib.warnings.warn(
                    '\n'.join(
                        (
                            Constants.SILENCE_MSG,
                            *[str(a) for a in args],
                            )
                        )
                    )
            else:
                lib.warnings.warn(Constants.WARN_MSG, stacklevel=1)
                _print(*args, **kwargs)

        __builtins__['print'] = _reprint

        self.log = fqr.log
        self.msg = mocking.examples.Pet
        return super().setUp()

    def test_01_print(self):
        """Test first print + warning interception."""

        msg = self.msg
        warn_msg = '\n'.join(
            (
                '.py:line_no: WarningClass: ',
                Constants.SILENCE_MSG,
                str(msg),
                'warnings.warn()'
                )
            )
        level = lib.logging.WARNING
        parsed = fqr.loggers.utl.parse_incoming_log_message(warn_msg, level)
        with self.assertLogs(self.log, level) as logger:
            print(msg)
            expected_output = lib.textwrap.indent(
                lib.json.dumps(
                    parsed,
                    default=fqr.core.strings.utl.convert_for_repr,
                    indent=Constants.INDENT,
                    sort_keys=True
                    ),
                Constants.INDENT * ' '
                )
            self.assertEqual(
                logger.records[0].msg,
                expected_output
                )

    def test_02_print(self):
        """Test subsequent, same message print interceptions."""

        msg = self.msg
        warn_msg = '\n'.join(
            (
                Constants.SILENCE_MSG,
                str(msg),
                )
            )
        with self.assertWarns(Warning) as emitter:
            with self.assertNoLogs(self.log):
                print(msg)
                self.assertEqual(
                    emitter.warnings[0].message.args[0],
                    warn_msg
                    )

    def test_03_print(self):
        """Test subsequent, new message print + warning interception."""

        msg = 'testing'
        warn_msg = '\n'.join(
            (
                '.py:line_no: WarningClass: ',
                Constants.SILENCE_MSG,
                str(msg),
                'warnings.warn()'
                )
            )
        level = lib.logging.WARNING
        parsed = fqr.loggers.utl.parse_incoming_log_message(warn_msg, level)
        with self.assertLogs(self.log, level) as logger:
            print(msg)
            expected_output = lib.textwrap.indent(
                lib.json.dumps(
                    parsed,
                    default=fqr.core.strings.utl.convert_for_repr,
                    indent=Constants.INDENT,
                    sort_keys=True
                    ),
                Constants.INDENT * ' '
                )
            self.assertEqual(
                logger.records[0].msg,
                expected_output
                )

    def test_04_ext_warning(self):
        """Test external warning interception."""

        msg = repr(self.msg)
        level = lib.logging.WARNING
        parsed = fqr.loggers.utl.parse_incoming_log_message(msg, level)
        with self.assertLogs(self.log, level) as logger:
            lib.warnings.warn(msg, stacklevel=1)
            expected_output = lib.textwrap.indent(
                lib.json.dumps(
                    parsed,
                    default=fqr.core.strings.utl.convert_for_repr,
                    indent=Constants.INDENT,
                    sort_keys=True
                    ),
                Constants.INDENT * ' '
                )
            self.assertEqual(
                logger.records[0].msg,
                expected_output
                )

    def test_05_invalid_log_type_exc(self):
        """Test correct exc is raised."""

        with self.assertNoLogs(self.log):
            self.assertRaises(
                fqr.loggers.exc.InvalidLogMessageTypeError,
                lambda: fqr.log.info(42)
                )

    def test_06_log_truncation(self):
        """Test long strings correctly truncated."""

        msg = {'longString': 'ABCDEFG' * 1024}
        level = lib.logging.CRITICAL
        with self.assertLogs(self.log, level) as logger:
            fqr.log.critical(msg)
            self.assertIn(
                fqr.core.strings.cfg.Constants.M_LINE_TOKEN,
                logger.records[0].msg
                )

    def test_07_log_redaction(self):
        """Test sensitive strings correctly redacted."""

        msg = {'apiKey': 'ABCDEFG'}
        level = lib.logging.INFO
        with self.assertLogs(self.log, level) as logger:
            fqr.log.info(msg)
            self.assertIn(
                '[ REDACTED :: API_KEY_TOKEN ]',
                logger.records[0].msg
                )

    def tearDown(self) -> None:
        Constants.ENV = 'local'
        return super().tearDown()


class TestTraceBackLogger(unittest.TestCase):
    """Fixture for testing logger with error tracebacks."""

    def setUp(self) -> None:
        Constants.LOG_TRACEBACK = True

        def _monkey_log(
            level: (
                lib.t.Literal[0]
                | lib.t.Literal[10]
                | lib.t.Literal[20]
                | lib.t.Literal[30]
                | lib.t.Literal[40]
                | lib.t.Literal[50]
                ),
            msg: lib.t.Any,
            args: 'lib.logging._ArgsType',
            exc_info: 'lib.logging._ExcInfoType' = True,
            extra: lib.t.Union[lib.t.Mapping[str, object], None] = None,
            stack_info: bool = False,
            stacklevel: int = 1,
            **kwargs: lib.t.Any
            ) -> None:
            """Monkey patch for `logger._log`."""

            sinfo = None
            if lib.logging._srcfile:  # pragma: no cover
                try:
                    fn, lno, func, sinfo = fqr.log.findCaller(
                        stack_info,
                        stacklevel
                        )
                except ValueError:
                    fn, lno, func = "(unknown file)", 0, "(unknown function)"
            else:  # pragma: no cover
                fn, lno, func = "(unknown file)", 0, "(unknown function)"

            msg_dict = fqr.loggers.utl.parse_incoming_log_message(msg, level)

            if isinstance(exc_info, BaseException):
                exc_info = (type(exc_info), exc_info, exc_info.__traceback__)
            elif not isinstance(exc_info, tuple):
                exc_info = lib.sys.exc_info()
            if (
                Constants.LOG_TRACEBACK
                and level >= lib.logging.ERROR
                and exc_info[-1] is not None
                and not isinstance(exc_info[1], KeyboardInterrupt)
                ):
                if 'printed' in msg_dict:
                    msg_final = (
                        fqr.loggers.typ.LogRecordWithPrintAndTraceBack(
                            traceback=lib.traceback.format_exc(),
                            message=msg_dict['message'],
                            printed=msg_dict['printed']  # type: ignore[typeddict-item]
                            )
                        )
                else:
                    msg_final = fqr.loggers.typ.LogRecordWithTraceBack(
                        traceback=lib.traceback.format_exc(),
                        message=msg_dict['message']
                        )
            else:
                msg_final = msg_dict

            record = fqr.log.makeRecord(
                fqr.log.name,
                level,
                fn,
                lno,
                lib.textwrap.indent(
                    lib.json.dumps(
                        fqr.core.strings.utl.convert_for_repr(msg_final),
                        default=fqr.core.strings.utl.convert_for_repr,
                        indent=Constants.INDENT,
                        sort_keys=True
                        ),
                    Constants.INDENT * ' '
                    ),
                tuple(),
                None,
                func,
                extra,
                sinfo
                )
            fqr.log.handle(record)

        fqr.log._log = _monkey_log
        self.log = fqr.log
        self.msg_str = 'example'
        return super().setUp()

    def test_01_log(self):
        """Test exc logging."""

        msg = self.msg_str
        level = lib.logging.ERROR
        parsed = fqr.loggers.utl.parse_incoming_log_message(msg, level)
        with self.assertLogs(self.log, level) as logger:
            try:
                _ = 1 / 0
            except Exception:
                self.log.error(msg)
                with_tb = fqr.loggers.typ.LogRecordWithTraceBack(
                    message=parsed['message'],
                    traceback=lib.traceback.format_exc()
                    )
                expected_output = lib.textwrap.indent(
                    lib.json.dumps(
                        with_tb,
                        default=fqr.core.strings.utl.convert_for_repr,
                        indent=Constants.INDENT,
                        sort_keys=True
                        ),
                    Constants.INDENT * ' '
                    )
            self.assertEqual(logger.records[0].msg, expected_output)

    def test_02_log(self):
        """Test exc logging - no trace for < ERROR."""

        msg = self.msg_str
        level = lib.logging.DEBUG
        parsed = fqr.loggers.utl.parse_incoming_log_message(msg, level)
        with self.assertLogs(self.log, level) as logger:
            try:
                _ = 1 / 0
            except Exception as e:
                self.log.info(msg, exc_info=e)
                expected_output = lib.textwrap.indent(
                    lib.json.dumps(
                        parsed,
                        default=fqr.core.strings.utl.convert_for_repr,
                        indent=Constants.INDENT,
                        sort_keys=True
                        ),
                    Constants.INDENT * ' '
                    )
            self.assertEqual(logger.records[0].msg, expected_output)

    def tearDown(self) -> None:
        Constants.LOG_TRACEBACK = False
        return super().tearDown()



"""Strings utility functions."""

__all__ = (
    'camel_case_to_snake_case',
    'cname_for',
    'convert_for_repr',
    'is_snake_case_iterable',
    'is_snake_case_string',
    'is_valid_datetime_str',
    'is_valid_number_str',
    'isCamelCaseIterable',
    'isCamelCaseString',
    'redact_key_value_pair',
    'redact_string',
    'snake_case_to_camel_case',
    'validate_casing',
    )

from .. import codecs

from . import cfg
from . import enm
from . import exc
from . import lib
from . import obj
from . import typ


class Constants(cfg.Constants):
    """Constants specific to this file."""

    CACHED_CNAMES: dict[tuple[typ.AnyString, tuple[typ.AnyString, ...]], typ.AnyString] = {}


def isCamelCaseString(
    string: str
    ) -> lib.t.TypeGuard[typ.string[typ.camelCase]]:
    """
    Check if `string` is valid `camelCase`.

    ---

    Checks for strict [lower] `camelCase` (i.e. `RESTful casing`) \
    according to the [Google Java Style Guide](https://google.github.io/styleguide/javaguide.html#s5.3-camel-case).

    Unlike Google, does *NOT* allow for an optional uppercase character \
    at the end of the `string`.

    """

    return _isCamelCaseString(string)


@lib.functools.cache
def _isCamelCaseString(string: str) -> bool:
    return bool(obj.Pattern.camelCase.match(string))


def is_snake_case_string(
    string: typ.AnyString
    ) -> lib.t.TypeGuard[typ.string[typ.snake_case]]:
    """
    Check if `string` is valid `snake_case`.

    ---

    Checks for strict [lower] `snake_case` (i.e. \
    `python attribute casing`).

    """

    return _is_snake_case_string(string)


@lib.functools.cache
def _is_snake_case_string(string: str) -> bool:
    return bool(obj.Pattern.snake_case.match(string))


@lib.functools.cache
def validate_casing(
    value: lib.t.Any,
    casing: typ.Casing
    ) -> lib.t.Optional[lib.Never]:
    """
    Assert value is `str` and of correct `Casing`.

    ---

    Raises `TypeError` if not `str`.

    Raises `StringCasingError` if `str` of incorrect `Casing`.

    """

    if not isinstance(value, str):
        raise TypeError(
            f'{value!s} is not a valid `str`.'
            )
    elif (
        casing == enm.SupportedCasing.snake_case.value
        and not is_snake_case_string(value)
        ):
        raise exc.StringCasingError(value, casing)
    elif (
        casing == enm.SupportedCasing.camelCase.value
        and not isCamelCaseString(value)
        ):
        raise exc.StringCasingError(value, casing)
    else:
        return None


@lib.functools.cache
def snake_case_to_camel_case(
    snake_case_string: typ.string[typ.snake_case]
    ) -> typ.string[typ.camelCase]:
    """Convert a valid `str[snake_case]` to `str[camelCase]`."""

    camelCaseString: typ.string[typ.camelCase] = (
        obj.Pattern.SnakeToCamelReplacements.sub(
            lambda match: match.group()[-1].upper(),
            snake_case_string
            )
        )

    return camelCaseString


@lib.functools.cache
def camel_case_to_snake_case(
    camelCaseString: typ.string[typ.camelCase]
    ) -> typ.string[typ.snake_case]:
    """Convert a valid `str[camelCase]` to `str[snake_case]`."""

    snake_case_string: typ.string[typ.snake_case] = (
        obj.Pattern.CamelToSnakeReplacements.sub(
            lambda match: '_' + match.group().lower(),
            camelCaseString
            )
        )

    return snake_case_string


@lib.t.overload
def is_snake_case_iterable(
    strings: 'typ.Mapping[str, typ.AnyType]'
    ) -> lib.t.TypeGuard[
        'typ.Mapping[typ.string[typ.snake_case], typ.AnyType]'
        ]: ...
@lib.t.overload
def is_snake_case_iterable(
    strings: lib.t.Iterable[str]
    ) -> 'lib.t.TypeGuard[lib.t.Iterable[typ.string[typ.snake_case]]]': ...
def is_snake_case_iterable(
    strings: 'lib.t.Iterable[str] | typ.Mapping[str, typ.AnyType]'
    ) -> lib.t.TypeGuard[
        lib.t.Union[
            'lib.t.Iterable[typ.string[typ.snake_case]]',
            'typ.Mapping[typ.string[typ.snake_case], typ.AnyType]'
            ]
        ]:
    """
    Check if all `strings` are `str[snake_case]`.

    ---

    Ignores leading and / or trailing underscores.

    """

    return all(
        is_snake_case_string(string)
        for _string
        in strings
        if (string := _string.strip('_'))
        )


@lib.t.overload
def isCamelCaseIterable(
    strings: 'typ.Mapping[str, typ.AnyType]'
    ) -> lib.t.TypeGuard[
        'typ.Mapping[typ.string[typ.camelCase], typ.AnyType]'
        ]: ...
@lib.t.overload
def isCamelCaseIterable(
    strings: lib.t.Iterable[str]
    ) -> 'lib.t.TypeGuard[lib.t.Iterable[typ.string[typ.camelCase]]]': ...
def isCamelCaseIterable(
    strings: 'lib.t.Iterable[str] | typ.Mapping[str, typ.AnyType]'
    ) -> lib.t.TypeGuard[
        lib.t.Union[
            'lib.t.Iterable[typ.string[typ.camelCase]]',
            'typ.Mapping[typ.string[typ.camelCase], typ.AnyType]'
            ]
        ]:
    """
    Check if all `strings` are `str[camelCase]`.

    ---

    Ignores leading and / or trailing underscores.

    """

    return all(
        isCamelCaseString(string)
        for _string
        in strings
        if (string := _string.strip('_'))
        )


@lib.t.overload
def cname_for(
    string: 'typ.AnyString | str',
    container: tuple[typ.string[typ.StringType], ...]
    ) -> lib.t.Optional[typ.string[typ.StringType]]: ...
@lib.t.overload
def cname_for(
    string: 'typ.AnyString | str',
    container: tuple[str, ...]
    ) -> lib.t.Optional[str]: ...
def cname_for(
    string: 'typ.AnyString | str',
    container: tuple[typ.string[typ.StringType] | str, ...]
    ) -> lib.t.Optional[typ.string[typ.StringType] | str]:
    """
    Get the actual, canonical name for valid `string`, as contained in \
    an arbitrary, valid, immutable `Container[str]` (i.e. a \
    `tuple[str, ...]`), agnostic of `string` casing and / or underscores.

    ---

    ### Example Usage

    ```py
    d = {
        '_id': 123,
        '_meaning_of_life': 42
        }

    cname_for(d, 'id')
    '_id'

    cname_for(d, 'meaningOfLife')
    42

    ```

    """

    v: typ.string[typ.StringType]
    if (__key := (string, container)) in Constants.CACHED_CNAMES:
        v = Constants.CACHED_CNAMES[__key]
    elif (
        (
            (k := (__k := string.strip('_'))) in container
            or (k := '_' + __k) in container
            or (k := __k + '_') in container
            or (k := '_' + __k + '_') in container
            )
        or (
            isCamelCaseString(__k)
            and (
                (
                    k := (_k := camel_case_to_snake_case(__k.strip('_')))
                    ) in container
                or (k := '_' + _k) in container
                or (k := _k + '_') in container
                or (k := '_' + _k + '_') in container
                )
            )
        or (
            is_snake_case_string(__k)
            and (
                (
                    k := (_k := snake_case_to_camel_case(__k.strip('_')))
                    ) in container
                or (k := '_' + _k) in container
                or (k := _k + '_') in container
                or (k := '_' + _k + '_') in container
                )
            )
        ):
        v = k
        Constants.CACHED_CNAMES[__key] = v
    else:
        v = None
        Constants.CACHED_CNAMES[__key] = v

    return v


def is_valid_number_str(
    any_str: str
    ) -> lib.t.TypeGuard[typ.string[typ.numeric]]:
    """
    Return `True` if python `str` is parsable as a valid \
    `numbers.Number`.

    """

    return bool(obj.Pattern.Number.match(any_str))


def is_valid_datetime_str(
    any_str: str
    ) -> lib.t.TypeGuard[typ.string[typ.datetime]]:
    """
    Return `True` if python `str` is parsable as a valid \
    `datetime`.

    """

    return bool(obj.Pattern.DateTime.match(any_str))


def redact_key_value_pair(key: str, value: str) -> str:
    """
    Redact potentially sensitive key, value pairs.

    ---

    Returns only the value (or a redacted version).

    """

    for id_, pattern in obj.KeyValueRedactionPatterns.items():
        if pattern.search(key) is not None:
            return f'[ REDACTED :: {id_} ]'

    return value


def redact_string(string: str) -> str:
    """Redact potentially sensitive values."""

    for id_, pattern in obj.RedactionPatterns.items():
        string = pattern.sub(
            repl=f'[ REDACTED :: {id_} ]',
            string=string
            )

    return string


@lib.t.overload
def convert_for_repr(
    obj_: 'typ.Mapping[typ.Primitive, typ.Serial]'
    ) -> dict[typ.Primitive, typ.Serial]: ...
@lib.t.overload
def convert_for_repr(obj_: typ.Primitive) -> typ.Primitive: ...
@lib.t.overload
def convert_for_repr(obj_: typ.Serial | lib.t.Any) -> typ.Serial: ...
def convert_for_repr(obj_: lib.t.Any) -> typ.Serial:
    """Recursively prepare `obj_` for neat `__repr__`."""

    if isinstance(obj_, str):
        redacted = redact_string(obj_)
        if len(redacted) > Constants.MAX_CHARS:
            return [Constants.M_LINE_TOKEN, *obj.StringWrapper.wrap(redacted)]
        else:
            return redacted
    elif obj_ is None or typ.utl.check.is_primitive(obj_):
        return obj_
    elif typ.utl.check.is_field(obj_):
        from ... import objects
        return repr(objects.typ.Field(obj_, obj_.type_))  # type: ignore[call-overload]
    elif typ.utl.check.is_typed(obj_):
        return {
            convert_for_repr(k): (
                convert_for_repr(redact_key_value_pair(k, v))
                if (
                    isinstance(k, str)
                    and isinstance((v := getattr(obj_, k)), str)
                    )
                else convert_for_repr(v)
                )
            for k
            in typ.utl.hint.collect_annotations(obj_)
            }
    elif typ.utl.check.is_mapping(obj_):
        return {
            convert_for_repr(k): (
                convert_for_repr(redact_key_value_pair(k, v))
                if (isinstance(k, str) and isinstance(v, str))
                else convert_for_repr(v)
                )
            for k, v
            in obj_.items()
            }
    elif typ.utl.check.is_array(obj_):
        if len(obj_) >= 1 and not all(isinstance(v, str) for v in obj_):
            return [
                convert_for_repr(v)
                for v
                in obj_
                ]
        else:
            return list(obj_)
    elif isinstance(obj_, (lib.types.FunctionType, lib.types.MethodType)):
        rtn_tp: typ.Serial | lib.t.Any = (
            obj_.__annotations__.get('return', codecs.utl.encode(obj_))
            )
        return convert_for_repr(rtn_tp)
    else:
        return codecs.utl.encode(obj_)



"""Loggers exceptions module."""

from .. import core

__all__ = (
    'InvalidLogMessageTypeError',
    *core.exc.__all__
    )

from .. core . exc import *

from . import cfg
from . import lib


class Constants(cfg.Constants):
    """Constant values specific to this file."""


class InvalidLogMessageTypeError(BasePackageException[lib.t.Any]):
    """
    Error raised when a log message of invalid data type is passed.

    """

    def __init__(self, message: lib.t.Any):
        super().__init__(
            ' '.join(
                (
                    'To avoid accidental log pollution,'
                    'fqr can only log: `dict, list, str, Object` types.',
                    f"The following message of type: '{type(message)!s}'",
                    f'was passed.\nmessage: {message!s}'
                    )
                ),
            message
            )



"""Loggers objects."""

__all__ = (
    'log',
    )

from .. import core

from . import cfg
from . import lib
from . import typ
from . import utl


class Constants(cfg.Constants):
    """Constant values specific to this file."""


lib.logging.Formatter.converter = lib.time.gmtime
lib.logging.Formatter.default_time_format = Constants.FTIME_LOG
lib.logging.Formatter.default_msec_format = Constants.FTIME_LOG_MSEC
lib.logging.basicConfig(
    format=(' ' * Constants.INDENT).join(
        (
            '{\n',
            '"level": %(levelname)s,\n',
            '"time": %(asctime)s,\n',
            '"log": %(name)s,\n',
            '"data": %(message)s\n}',
            )
        ),
    )

log = lib.logging.getLogger(__name__)
"""
Centralized application log.

Pre-configured so you do not need to. By default, this loggger \
will do everything it can to keep your log stream as actionable \
and free from pollution as possible.

* emits neatly formatted JSON log messages
* intercepts forgotten print statements
    \
    * silences them in all deployed environments
* intercepts forgotten debug-level logs
    \
    * silences them in deployed environments > dev
* intercepts irritating warnings, displaying them once as \
neatly formatted warning level log messages
* automatically redacts things that should never be logged \
in the first place (like API Keys, credentials, SSN's, etc.)

---

Usage
-----

The expectation is this will be the only log used across \
an application.

* Set logging level through the `LOG_LEVEL` environment variable.
    \
    * Defaults to 'DEBUG' if `Constants.ENV` is either `local` (default) \
    or `dev`, otherwise 'INFO'.

---

Special Rules
-------------

* Can only log `str`, `dict`, and `Object` types.

* Automatically redacts almost all sensitive data, including \
api keys, tokens, credit card numbers, connection strings, \
and secrets.

* All `warnings` will be filtered through this log and \
displayed only once.

* All `print` statements will be silenced *except* when \
`Constants.ENV` is set to 'local' (its default if `ENV` \
is unavailable in `os.environ` at runtime).
    \
    * Best practice is to set log level to 'DEBUG' and \
    use the `log.debug` method in place of `print` statements.
    \
    * `warnings` will be displayed once for all `print` \
    statements that would otherwise be silenced in any \
    non-local development environment.

---

Usage Examples
--------------

```py
import fgr

fgr.log.debug('example')
# >>>
# {
#   "level": DEBUG,
#   "time": 2024-02-25 15:30:01.061 UTC,
#   "log": fgr.core.log,
#   "data":   {
#     "message": "example"
#   }
# }

fgr.log.info({'str': 'example', 'a': 2})
# >>>
# {
#   "level": INFO,
#   "time": 2024-02-25 15:31:11.118 UTC,
#   "log": fgr.core.log,
#   "data":   {
#     "a": 2,
#     "str": "example"
#   }
# }


class Pet(fgr.Object):
    \"""A pet.\"""

    id_: fgr.Field[str]
    _alternate_id: fgr.Field[str]

    name: fgr.Field[str]
    type: fgr.Field[str]
    in_: fgr.Field[str]
    is_tail_wagging: fgr.Field[bool] = True


fgr.log.debug(Pet)
# >>>
# {
#   "level": DEBUG,
#   "time": 2024-02-25 15:30:01.339 UTC,
#   "log": fgr.core.log,
#   "data":   {
#     "Pet": {
#       "_alternate_id": "Field[str]",
#       "id": "Field[str]",
#       "in": "Field[str]",
#       "is_tail_wagging": "Field[bool]",
#       "name": "Field[str]",
#       "type": "Field[str]"
#     }
#   }
# }

fgr.log.debug(
    Pet(
        id_='abc1234',
        name='Fido',
        type='dog',
        )
    )
# >>>
# {
#   "level": DEBUG,
#   "time": 2024-02-25 15:30:01.450 UTC,
#   "log": fgr.core.log,
#   "data":   {
#     "Pet": {
#       "_alternate_id": null,
#       "id": "abc1234",
#       "in": null,
#       "is_tail_wagging": true,
#       "name": "Fido",
#       "type": "dog"
#     }
#   }
# }

```

"""

log.setLevel(lib.logging._nameToLevel[Constants.LOG_LEVEL])

lib.warnings.simplefilter('once')
lib.logging.captureWarnings(True)
lib.logging.Logger.manager.loggerDict['py.warnings'] = log

if not Constants.LOG_PRINTS:

    _print = __builtins__['print']  # type: ignore[index]

    def _reprint(*args: lib.t.Any, **kwargs: lib.t.Any) -> None:
        if Constants.ENV in Constants.DEPLOY_ENVS:  # pragma: no cover (still covered)
            lib.warnings.warn(
                '\n'.join(
                    (
                        Constants.SILENCE_MSG,
                        *[str(a) for a in args],
                        )
                    )
                )
        else:
            lib.warnings.warn(Constants.WARN_MSG, stacklevel=1)
            _print(*args, **kwargs)

    __builtins__['print'] = _reprint


def _monkey_log(
    level: (
        lib.t.Literal[0]
        | lib.t.Literal[10]
        | lib.t.Literal[20]
        | lib.t.Literal[30]
        | lib.t.Literal[40]
        | lib.t.Literal[50]
        ),
    msg: lib.t.Any,
    args: 'lib.logging._ArgsType',
    exc_info: 'lib.logging._ExcInfoType' = True,
    extra: lib.t.Union[lib.t.Mapping[str, object], None] = None,
    stack_info: bool = False,
    stacklevel: int = 1,
    **kwargs: lib.t.Any
    ) -> None:
    """Monkey patch for `logger._log`."""

    sinfo = None
    if lib.logging._srcfile:  # pragma: no cover
        try:
            fn, lno, func, sinfo = log.findCaller(stack_info, stacklevel)
        except ValueError:
            fn, lno, func = "(unknown file)", 0, "(unknown function)"
    else:  # pragma: no cover
        fn, lno, func = "(unknown file)", 0, "(unknown function)"

    if msg == '%s' and args and isinstance(args, tuple):  # pragma: no cover
        msg_ = args[0]
    else:
        msg_ = msg

    msg_dict = utl.parse_incoming_log_message(msg_, level)

    if isinstance(exc_info, BaseException):
        exc_info = (type(exc_info), exc_info, exc_info.__traceback__)
    elif not isinstance(exc_info, tuple):
        exc_info = lib.sys.exc_info()
    if (
        Constants.LOG_TRACEBACK
        and level >= lib.logging.ERROR
        and any(exc_info)
        and not isinstance(exc_info[1], KeyboardInterrupt)
        ):
        if 'printed' in msg_dict:  # pragma: no cover (still covered)
            msg_final = typ.LogRecordWithPrintAndTraceBack(
                message=msg_dict['message'],
                printed=msg_dict['printed'],  # type: ignore[typeddict-item]
                traceback=lib.traceback.format_exc()
                )
        else:  # pragma: no cover (still covered)
            msg_final = typ.LogRecordWithTraceBack(
                message=msg_dict['message'],
                traceback=lib.traceback.format_exc()
                )
    else:
        msg_final = msg_dict

    record = log.makeRecord(
        log.name,
        level,
        fn,
        lno,
        lib.textwrap.indent(
            lib.json.dumps(
                core.strings.utl.convert_for_repr(msg_final),
                default=core.strings.utl.convert_for_repr,
                indent=Constants.INDENT,
                sort_keys=True
                ),
            Constants.INDENT * ' '
            ),
        tuple(),
        None,
        func,
        extra,
        sinfo
        )
    log.handle(record)


log._log = _monkey_log



import unittest


class TestReadMe(unittest.TestCase):
    """Test README.md advanced usage example."""

    def test_01_readme(self):
        """Test readme."""

        # START README.md

        try:
            import fqr

            class Flea(fqr.Object):
                """A nuisance."""

                name: fqr.Field[str] = 'FLEA'


            class Pet(fqr.Object):
                """A pet."""

                id_: fqr.Field[str]
                _alternate_id: fqr.Field[int]

                name: fqr.Field[str]
                type_: fqr.Field[str] = {
                    'default': 'dog',
                    'enum': ['cat', 'dog'],
                    'required': True,
                    }

                in_: fqr.Field[str]
                is_tail_wagging: fqr.Field[bool] = fqr.Field(
                    default=True,
                    enum=[True, False],
                    required=True,
                    )

                fleas: fqr.Field[list[Flea]] = [
                    Flea(name='flea1'),
                    Flea(name='flea2')
                    ]


            # Automatic case handling.
            request_body = {
                'id': 'abc123',
                'alternateId': 123,
                'name': 'Bob',
                'type': 'dog',
                'in': 'timeout',
                'isTailWagging': False
                }
            pet = Pet(request_body)

            assert pet['alternate_id'] == pet._alternate_id == request_body['alternateId']
            assert dict(pet) == {k: v for k, v in pet.items()} == pet.to_dict()

            # Automatic, mutation-safe "default factory".
            dog = Pet(id='abc321', alternate_id=321, name='Fido')
            assert pet.fleas[0] is not dog.fleas[0]

            # Automatic memory optimization.
            assert Flea().__sizeof__() == (len(Flea.__slots__) * 8) + 16 == 24

            class Flet(Flea, Pet):
                ...

            class Pea(Pet, Flea):
                ...

            assert Flet().__sizeof__() == (len(Flet.__base__.__slots__) * 8) + 16 == 72
            assert Pea().__sizeof__() == (len(Pea.__base__.__slots__) * 8) + 16 == 72
            assert Flet().name == 'FLEA' != Pea().name

            # Intuitive, database agnostic query generation.
            assert isinstance(Pet.is_tail_wagging, fqr.Field)
            assert isinstance(Pet.type_, fqr.Field)

            assert dog.type_ == Pet.type_.default == 'dog'

            query = (
                (
                    (Pet.type_ == 'dog')
                    & (Pet.name == 'Fido')
                    )
                | Pet.name % ('fido', 0.75)
                )
            query += 'name'
            assert dict(query) == {
                'limit': None,
                'or': [
                    {
                        'and': [
                            {
                                'eq': 'dog',
                                'field': 'type',
                                'limit': None,
                                'sorting': []
                                },
                            {
                                'eq': 'Fido',
                                'field': 'name',
                                'limit': None,
                                'sorting': []
                                }
                            ],
                        'limit': None,
                        'sorting': []
                        },
                    {
                        'field': 'name',
                        'like': 'fido',
                        'limit': None,
                        'sorting': [],
                        'threshold': 0.75
                        }
                    ],
                'sorting': [
                    {
                        'direction': 'asc',
                        'field': 'name'
                        }
                    ]
                }

            # END README.md

        except AssertionError:
            self.assertTrue(False)
        else:
            self.assertTrue(True)



__all__ = (
    'examples',
    'Derivative',
    'DubDeriv',
    'MixinDeriv',
    'NewDeriv',
    'TripDeriv',
    'AntiTripDeriv',
    )

import fqr

from fqr . core import lib

from . import examples


class Derivative(fqr.Object):
    """Simple test derivative."""

    secondary_key: fqr.Field[int] = 123
    str_field: fqr.Field[str] = 'abc'
    bool_field: fqr.Field[bool] = True
    int_field: fqr.Field[int] = 2
    forward_ref_alias_field: 'fqr.Field[fqr.core.typ.AnyType]' = 2
    forward_ref_union_field: 'fqr.Field[float | int | tuple[int | float, ...]]' = 2  # noqa
    forward_ref_field: fqr.Field[list['Derivative']] = []
    enumerated_bool_field: fqr.Field[bool] = fqr.Field(
        default=False,
        enum=fqr.core.enm.Boolean,
        )
    from_dict_field: fqr.Field[lib.t.Optional[str]] = {
        'default': 'asc',
        'enum': {'asc', 'desc'},
        'required': False,
        'type': lib.t.Optional[str]
        }
    null_field: fqr.Field[fqr.core.typ.NoneType] = None
    non_nullable_field: fqr.Field[int] = fqr.Field(default=4, type=int)
    required_field: fqr.Field[int]
    date_field: fqr.Field[lib.datetime.date] = (  # noqa: E731
        lambda: lib.datetime.datetime.now(lib.datetime.timezone.utc).date()
        )
    datetime_field: fqr.Field[lib.datetime.datetime] = (  # noqa: E731
        lambda: lib.datetime.datetime.now(lib.datetime.timezone.utc)
        )
    decimal_field: fqr.Field[lib.decimal.Decimal] = lib.decimal.Decimal(1e-3)
    tuple_field: fqr.Field[tuple] = (1, 2)
    generic_tuple_field: fqr.Field[tuple[str, float, bool]] = ('a', 2.5, False)  # noqa


class DubDeriv(Derivative):

    test_again: fqr.Field[bool] = True
    bob: fqr.Field[str] = 'Dan'
    other_field: fqr.Field[lib.t.Optional[str]] = fqr.Field(
        default='Paul',
        enum=['Paul'],
        type_=lib.t.Optional[str]
        )

    def do_stuff(self):
        ...


class MixinDeriv(fqr.Object):

    test_again: fqr.Field[bool] = True
    bob: fqr.Field[str] = 'David'
    other_field: fqr.Field[lib.t.Optional[str]] = fqr.Field(
        default='Albert',
        enum=['Albert'],
        type_=lib.t.Optional[str]
        )

    def do_stuff(self):
        ...


class NewDeriv(fqr.Object):

    anti_field_1: fqr.Field[str] = 'cba'
    anti_field_2: fqr.Field[bool] = False
    generic_tuple_deriv_field: fqr.Field[tuple[MixinDeriv, ...]] = lambda: (  # noqa: E731
        MixinDeriv(bob='Frank'),
        MixinDeriv(bob='Bob'),
        )


class TripDeriv(MixinDeriv, DubDeriv):

    test_another: fqr.Field[bool] = False
    new_deriv: fqr.Field[NewDeriv] = NewDeriv()
    dict_field: fqr.Field[dict] = {'record_id': 'Arnold'}
    generic_dict_field: fqr.Field[dict[str, float]] = {'record_id': 1.23}


class AntiTripDeriv(DubDeriv, MixinDeriv):

    test_another: fqr.Field[bool] = False
    new_deriv: fqr.Field[NewDeriv] = NewDeriv()
    dict_field: fqr.Field[dict] = {'record_id': 'Lauren'}
    generic_dict_field: fqr.Field[dict[str, float]] = {'record_id': 1.23}



"""Strings enumerations."""

from .. import enm

__all__ = (
    'SupportedCasing',
    *enm.__all__
    )

from .. enm import *

from . import lib
from . import typ


class SupportedCasing(lib.enum.Enum):
    """Valid string casings."""

    camelCase  = typ.camelCase
    snake_case = typ.snake_case



import unittest

import fqr

from ... import mocking


class TestQuery(unittest.TestCase):
    """Fixture for testing the object."""

    def setUp(self) -> None:
        self.cls = mocking.Derivative
        return super().setUp()

    def test_01_eq(self):
        """Test __eq__."""

        self.assertIsInstance(
            self.cls.str_field == 'abc',
            fqr.objects.queries.EqQueryCondition
            )

    def test_02_ne(self):
        """Test __ne__."""

        self.assertIsInstance(
            self.cls.bool_field != True,  # noqa: E712
            fqr.objects.queries.NeQueryCondition
            )

    def test_03_le(self):
        """Test __le__."""

        self.assertIsInstance(
            self.cls.int_field <= 1,
            fqr.objects.queries.LeQueryCondition
            )

    def test_04_lt(self):
        """Test __lt__."""

        self.assertIsInstance(
            self.cls.int_field < 1,
            fqr.objects.queries.LtQueryCondition
            )

    def test_05_ge(self):
        """Test __ge__."""

        self.assertIsInstance(
            self.cls.int_field >= 1,
            fqr.objects.queries.GeQueryCondition
            )

    def test_06_gt(self):
        """Test __gt__."""

        self.assertIsInstance(
            self.cls.int_field > 1,
            fqr.objects.queries.GtQueryCondition
            )

    def test_07_contains(self):
        """Test __lshift__."""

        self.assertIsInstance(
            self.cls.str_field << 't',
            fqr.objects.queries.ContainsQueryCondition
            )

    def test_08_contains_error(self):
        """Test __lshift__ error."""

        self.assertRaises(
            fqr.objects.exc.InvalidContainerComparisonTypeError,
            lambda: self.cls.int_field << 't',
            )

    def test_09_similar(self):
        """Test __mod__."""

        self.assertIsInstance(
            self.cls.str_field % 't',
            fqr.objects.queries.SimilarQueryCondition
            )

    def test_10_similar_with_threshold(self):
        """Test __mod__ with threshold."""

        self.assertEqual(
            self.cls.str_field % ('t', 0.75),
            fqr.objects.queries.SimilarQueryCondition(
                field=self.cls.str_field.name,
                like='t',
                threshold=0.75
                )
            )

    def test_11_and_query(self):
        """Test __and__."""

        self.assertEqual(
            (
                (self.cls.int_field >= 1)
                & (self.cls.int_field < 10)
                ),
            fqr.objects.queries.AndQuery(
                and_=[
                    self.cls.int_field >= 1,
                    self.cls.int_field < 10
                    ]
                )
            )

    def test_12_or_query(self):
        """Test __or__."""

        self.assertEqual(
            (
                (self.cls.int_field >= 1)
                | (self.cls.int_field < 10)
                ),
            fqr.objects.queries.OrQuery(
                or_=[
                    self.cls.int_field >= 1,
                    self.cls.int_field < 10
                    ]
                )
            )

    def test_13_invert_query(self):
        """Test __invert__."""

        self.assertEqual(
            ~(_q := self.cls.int_field >= 1),
            fqr.objects.queries.InvertQuery(invert=_q)
            )

    def test_14_sort_asc_query(self):
        """Test __iadd__."""

        q = self.cls.int_field >= 1
        q += self.cls.int_field.name
        self.assertEqual(
            q.sorting[0],
            fqr.objects.queries.QuerySortBy(
                field=self.cls.int_field.name,
                direction='asc'
                )
            )

    def test_15_sort_desc_query(self):
        """Test __isub__."""

        q = self.cls.int_field >= 1
        q -= self.cls.int_field.name
        self.assertEqual(
            q.sorting[0],
            fqr.objects.queries.QuerySortBy(
                field=self.cls.int_field.name,
                direction='desc'
                )
            )

    def test_16_similar_without_threshold(self):
        """Test __mod__ without threshold."""

        self.assertEqual(
            self.cls.str_field % ('t', 'test'),
            fqr.objects.queries.SimilarQueryCondition(
                field=self.cls.str_field.name,
                like='t',
                threshold=fqr.objects.enm.MatchThreshold.default.value
                )
            )



"""Module exceptions unit tests."""

import pickle
import unittest

import fqr

from . import cfg


class Constants(cfg.Constants):
    """Constant values specific to unit tests in this file."""


class TestExceptions(unittest.TestCase):
    """Fixture for testing."""

    def test_01_serialization(self):
        """Test multi-arg exc serializes correctly."""

        exc = fqr.core.strings.exc.StringCasingError(
            Constants.INVALID_STRING_CASING_EXAMPLE,
            fqr.core.strings.typ.camelCase
            )
        dump = pickle.dumps(exc)
        deserialized_exc: fqr.core.strings.exc.StringCasingError = (
            pickle.loads(dump)
            )
        self.assertTupleEqual(exc.args, deserialized_exc.args)



"""Static files for docs gen."""

__all__ = (
    'config_template',
    )

import os

_root = __file__.removesuffix('__init__.py')

with open(os.path.join(_root, 'conf.tpl'), 'r') as f:
    config_template = f.read()



"""Typing objects."""

__all__ = (
    'string',
    'ArrayProto',
    'FieldPattern',
    'MappingProto',
    'MetaLike',
    'ObjectLike',
    'SupportsAnnotations',
    'SupportsParams',
    'VariadicArrayProto',
    'WrapperPattern',
    )

from . import cfg
from . import lib

if lib.t.TYPE_CHECKING:  # pragma: no cover
    from . import typ  # noqa: F401


class Constants(cfg.Constants):
    """Constant values specific to this file."""


# Note: these need to be here to avoid circular import.
# They are however imported by adjacent typ and injected
# to typ's __all__ for consistency.
AnyType = lib.t.TypeVar('AnyType')
AnyOtherType = lib.t.TypeVar('AnyOtherType')
AnyTypeCo = lib.t.TypeVar('AnyTypeCo', covariant=True)
AnyOtherTypeCo = lib.t.TypeVar('AnyOtherTypeCo', covariant=True)
ArgsType = lib.TypeVarTuple('ArgsType')
StringType = lib.t.TypeVar('StringType', bound='typ.StringFormat')


class ArrayProto(lib.t.Protocol, lib.t.Collection[AnyTypeCo]):
    """Protocol for a generic, single-parameter array."""

    def __init__(
        self,
        iterable: lib.t.Iterable[AnyTypeCo],
        /
        ) -> None: ...

    def __iter__(self) -> lib.t.Iterator[AnyTypeCo]: ...


class VariadicArrayProto(
    ArrayProto[tuple[lib.Unpack[ArgsType]]],
    lib.t.Protocol
    ):
    """Protocol for a generic, any-parameter array."""

    def __hash__(self) -> int: ...


class MappingProto(
    lib.t.Protocol,
    lib.t.Generic[AnyTypeCo, AnyOtherTypeCo]
    ):
    """Protocol for a generic, double-parameter mapping."""

    def __init__(self, *args: lib.t.Any, **kwargs: lib.t.Any) -> None: ...

    def __iter__(self) -> lib.t.Iterator[AnyTypeCo]: ...

    def __getitem__(
        self,
        __name: str,
        __default: lib.t.Optional[AnyType] = None
        ) -> AnyTypeCo | AnyType: ...

    def items(self) -> lib.t.ItemsView[AnyTypeCo, AnyOtherTypeCo]: ...

    def keys(self) -> lib.t.KeysView[AnyTypeCo]: ...

    def values(self) -> lib.t.ValuesView[AnyOtherTypeCo]: ...


class SupportsAnnotations(lib.t.Protocol):
    """
    Protocol for a typed object.

    ---

    Typed objects include `dataclass`, `TypedDict`, `pydantic.Model`, \
    and both `fqr.Field` and `fqr.Object` amongst others.

    """

    __annotations__: dict[str, lib.t.Any]
    __bases__: tuple[type, ...]

    def __init__(self, *args: lib.t.Any, **kwargs: lib.t.Any) -> None: ...


class SupportsParams(lib.t.Protocol, lib.t.Generic[lib.Unpack[ArgsType]]):
    """Protocol for a generic with any number of parameters."""

    if lib.sys.version_info >= (3, 9):
        def __class_getitem__(
            cls,
            item: tuple[lib.Unpack[ArgsType]],
            /
            ) -> lib.types.GenericAlias: ...

    __args__: tuple[lib.Unpack[ArgsType]]

    def __hash__(self) -> int: ...


class MetaLike(lib.t.Protocol):
    """Meta protocol."""

    __annotations__: 'typ.SnakeDict'
    __dataclass_fields__: 'lib.t.ClassVar[typ.DataClassFields]'


class ObjectLike(lib.t.Protocol):
    """Object protocol."""

    __annotations__: 'typ.SnakeDict'
    __bases__: tuple[type, ...]
    __dataclass_fields__: 'lib.t.ClassVar[typ.DataClassFields]'

    def __contains__(self, __key: lib.t.Any, /) -> bool: ...

    def __getitem__(self, __key: lib.t.Any, /) -> lib.t.Any: ...

    def __setitem__(
        self,
        __key: str,
        __value: lib.t.Any
        ) -> lib.t.Optional[lib.Never]: ...

    def __ior__(self, other: 'ObjectLike', /) -> lib.Self: ...

    def get(
        self,
        __key: 'typ.AnyString',
        __default: AnyType = None
        ) -> lib.t.Any | AnyType: ...

    def items(
        self
        ) -> 'lib.t.ItemsView[typ.string[typ.snake_case], lib.t.Any]': ...

    @classmethod
    def keys(cls) -> 'lib.t.KeysView[typ.string[typ.snake_case]]': ...

    def pop(
        self,
        __key: str,
        /,
        __default: AnyType = Constants.UNDEFINED
        ) -> AnyType | lib.t.Any | lib.Never: ...

    def setdefault(
        self,
        __key: str,
        __value: lib.t.Any
        ) -> lib.t.Optional[lib.Never]: ...

    def update(self, other: 'ObjectLike', /) -> None: ...

    def values(self) -> lib.t.ValuesView[lib.t.Any]: ...

    @lib.t.overload
    def to_dict(
        self,
        camel_case: lib.t.Literal[False] = False,
        include_null: bool = True
        ) -> 'typ.SnakeDict': ...
    @lib.t.overload
    def to_dict(
        self,
        camel_case: lib.t.Literal[True],
        include_null: bool
        ) -> 'typ.CamelDict': ...
    @lib.t.overload
    def to_dict(
        self,
        camel_case: bool,
        include_null: bool
        ) -> 'typ.SnakeDict | typ.CamelDict': ...
    def to_dict(
        self,
        camel_case: bool = False,
        include_null: bool = True
        ) -> 'typ.SnakeDict | typ.CamelDict': ...


FieldPattern = lib.re.compile(
    r'(fqr(\.[a-zA-Z]{1,32}){0,32}\.)?Field'
    r'\[((\[)?[\.\|\,a-zA-Z0-9_ ]{1,64}(\])?){1,64}\]'
    )

WrapperPattern = lib.re.compile(
    r'([a-zA-Z]{1,64}\.?)?(Annotated|ClassVar|Final|InitVar)'
    r'\[((\[)?[\.\|\,a-zA-Z0-9_ ]{1,64}(\])?){1,64}\]'
    )


class string(str, lib.t.Generic[StringType]):
    """Generic `str` protocol."""

    @lib.t.overload  # type: ignore[no-overload-impl]
    def __new__(cls, object: object = ...) -> lib.Self: ...
    @lib.t.overload
    def __new__(
        cls,
        object: 'lib.builtins.ReadableBuffer',  # type: ignore[name-defined]
        encoding: str = ...,
        errors: str = ...
        ) -> lib.Self: ...
    @lib.t.overload  # type: ignore[no-overload-impl]
    def capitalize(self: lib.LiteralString) -> lib.LiteralString: ...
    @lib.t.overload
    def capitalize(self) -> 'string[StringType] | string[lib.t.Any]': ...  # type: ignore[misc]
    @lib.t.overload  # type: ignore[no-overload-impl]
    def casefold(self: lib.LiteralString) -> lib.LiteralString: ...
    @lib.t.overload
    def casefold(self) -> 'string[StringType] | string[lib.t.Any]': ...  # type: ignore[misc]
    @lib.t.overload  # type: ignore[no-overload-impl]
    def center(
        self: lib.LiteralString,
        width: lib.t.SupportsIndex,
        fillchar: lib.LiteralString = " ",
        /
        ) -> lib.LiteralString: ...
    @lib.t.overload
    def center(  # type: ignore[misc]
        self,
        width: lib.t.SupportsIndex,
        fillchar: str = " ",
        /
        ) -> 'string[StringType] | string[lib.t.Any]': ...
    @lib.t.overload  # type: ignore[no-overload-impl]
    def expandtabs(
        self: lib.LiteralString,
        tabsize: lib.t.SupportsIndex = 8
        ) -> lib.LiteralString: ...
    @lib.t.overload
    def expandtabs(  # type: ignore[misc]
        self,
        tabsize: lib.t.SupportsIndex = 8
        ) -> 'string[StringType] | string[lib.t.Any]': ...
    @lib.t.overload  # type: ignore[misc, no-overload-impl]
    def format(  # type: ignore[empty-body]
        self,
        *args: object,
        **kwargs: object
        ) -> 'string[StringType] | string[lib.t.Any]': ...
    def format_map(  # type: ignore[empty-body]
        self,
        mapping: 'lib.builtins._FormatMapMapping',
        /
        ) -> 'string[StringType] | string[lib.t.Any]': ...
    @lib.t.overload  # type: ignore[no-overload-impl]
    def join(
        self: lib.LiteralString,
        iterable: lib.t.Iterable[lib.LiteralString],
        /
        ) -> lib.LiteralString: ...
    @lib.t.overload
    def join(  # type: ignore[misc]
        self,
        iterable: 'lib.t.Iterable[string[StringType]]',
        /
        ) -> 'string[StringType]': ...
    @lib.t.overload
    def join(  # type: ignore[misc]
        self,
        iterable: str,
        /
        ) -> 'string[lib.t.Any]': ...
    @lib.t.overload  # type: ignore[no-overload-impl]
    def ljust(
        self: lib.LiteralString,
        width: lib.t.SupportsIndex,
        fillchar: lib.LiteralString = " ",
        /
        ) -> lib.LiteralString: ...
    @lib.t.overload
    def ljust(  # type: ignore[misc]
        self,
        width: lib.t.SupportsIndex,
        fillchar: str = " ",
        /
        ) -> 'string[StringType] | string[lib.t.Any]': ...
    @lib.t.overload  # type: ignore[no-overload-impl]
    def lower(self: lib.LiteralString) -> lib.LiteralString: ...
    @lib.t.overload
    def lower(self) -> 'string[StringType] | string[lib.t.Any]': ...  # type: ignore[misc]
    @lib.t.overload  # type: ignore[no-overload-impl]
    def lstrip(
        self: lib.LiteralString,
        chars: lib.LiteralString | None = None,
        /
        ) -> lib.LiteralString: ...
    @lib.t.overload
    def lstrip(  # type: ignore[misc]
        self,
        chars: str | None = None,
        /
        ) -> 'string[StringType] | string[lib.t.Any]': ...
    @lib.t.overload  # type: ignore[no-overload-impl]
    def partition(
        self: lib.LiteralString,
        sep: lib.LiteralString,
        /
        ) -> tuple[
            lib.LiteralString,
            lib.LiteralString,
            lib.LiteralString
            ]: ...
    @lib.t.overload
    def partition(self, sep: str, /) -> tuple[  # type: ignore[misc]
        'string[StringType] | string[lib.t.Any]', 
        'string[StringType] | string[lib.t.Any]', 
        'string[StringType] | string[lib.t.Any]'
        ]: ...
    if lib.sys.version_info >= (3, 13):  # pragma: no cover
        @lib.t.overload  # type: ignore[no-overload-impl]
        def replace(
            self: lib.LiteralString,
            old: lib.LiteralString,
            new: lib.LiteralString,
            /,
            count: lib.t.SupportsIndex = -1
        ) -> lib.LiteralString: ...
        @lib.t.overload
        def replace(  # type: ignore[misc]
            self,
            old: str,
            new: str,
            /,
            count: lib.t.SupportsIndex = -1
            ) -> 'string[StringType] | string[lib.t.Any]': ...
    else:  # pragma: no cover
        @lib.t.overload  # type: ignore[no-redef, no-overload-impl]
        def replace(
            self: lib.LiteralString,
            old: lib.LiteralString,
            new: lib.LiteralString,
            count: lib.t.SupportsIndex = -1,
            /
        ) -> lib.LiteralString: ...
        @lib.t.overload
        def replace(  # type: ignore[misc]
            self,
            old: str,
            new: str,
            count: lib.t.SupportsIndex = -1,
            /
            ) -> 'string[StringType] | string[lib.t.Any]': ...
    if lib.sys.version_info >= (3, 9):
        @lib.t.overload  # type: ignore[no-overload-impl]
        def removeprefix(
            self: lib.LiteralString,
            prefix: lib.LiteralString,
            /
            ) -> lib.LiteralString: ...
        @lib.t.overload
        def removeprefix(  # type: ignore[misc]
            self,
            prefix: str,
            /
            ) -> 'string[StringType] | string[lib.t.Any]': ...
        @lib.t.overload  # type: ignore[no-overload-impl]
        def removesuffix(
            self: lib.LiteralString,
            suffix: lib.LiteralString,
            /
            ) -> lib.LiteralString: ...
        @lib.t.overload
        def removesuffix(  # type: ignore[misc]
            self,
            suffix: str,
            /
            ) -> 'string[StringType] | string[lib.t.Any]': ...

    @lib.t.overload  # type: ignore[no-overload-impl]
    def rjust(
        self: lib.LiteralString,
        width: lib.t.SupportsIndex,
        fillchar: lib.LiteralString = " ",
        /
        ) -> lib.LiteralString: ...
    @lib.t.overload
    def rjust(  # type: ignore[misc]
        self,
        width: lib.t.SupportsIndex,
        fillchar: str = " ",
        /
        ) -> 'string[StringType] | string[lib.t.Any]': ...
    @lib.t.overload  # type: ignore[no-overload-impl]
    def rpartition(
        self: lib.LiteralString,
        sep: lib.LiteralString,
        /
        ) -> tuple[
            lib.LiteralString,
            lib.LiteralString,
            lib.LiteralString
            ]: ...
    @lib.t.overload
    def rpartition(  # type: ignore[misc]
        self,
        sep: str,
        /
        ) -> tuple[
            'string[StringType] | string[lib.t.Any]', 
            'string[StringType] | string[lib.t.Any]', 
            'string[StringType] | string[lib.t.Any]'
            ]: ...
    @lib.t.overload  # type: ignore[no-overload-impl]
    def rsplit(
        self: lib.LiteralString,
        sep: lib.LiteralString | None = None,
        maxsplit: lib.t.SupportsIndex = -1
        ) -> list[lib.LiteralString]: ...
    @lib.t.overload
    def rsplit(  # type: ignore[misc]
        self,
        sep: str | None = None,
        maxsplit: lib.t.SupportsIndex = -1
        ) -> 'list[string[StringType] | string[lib.t.Any]]': ...
    @lib.t.overload  # type: ignore[no-overload-impl]
    def rstrip(
        self: lib.LiteralString,
        chars: lib.LiteralString | None = None,
        /
        ) -> lib.LiteralString: ...
    @lib.t.overload
    def rstrip(  # type: ignore[misc]
        self,
        chars: str | None = None,
        /
        ) -> 'string[StringType] | string[lib.t.Any]': ...
    @lib.t.overload  # type: ignore[no-overload-impl]
    def split(
        self: lib.LiteralString,
        sep: lib.LiteralString | None = None,
        maxsplit: lib.t.SupportsIndex = -1
        ) -> list[lib.LiteralString]: ...
    @lib.t.overload
    def split(  # type: ignore[misc]
        self,
        sep: str | None = None,
        maxsplit: lib.t.SupportsIndex = -1
        ) -> 'list[string[StringType] | string[lib.t.Any]]': ...
    @lib.t.overload  # type: ignore[no-overload-impl]
    def splitlines(
        self: lib.LiteralString,
        keepends: bool = False
        ) -> list[lib.LiteralString]: ...
    @lib.t.overload
    def splitlines(  # type: ignore[misc]
        self,
        keepends: bool = False
        ) -> 'list[string[StringType] | string[lib.t.Any]]': ...
    @lib.t.overload  # type: ignore[no-overload-impl]
    def strip(
        self: lib.LiteralString,
        chars: lib.LiteralString | None = None,
        /
        ) -> lib.LiteralString: ...
    @lib.t.overload
    def strip(  # type: ignore[misc]
        self,
        chars: str | None = None,
        /
        ) -> 'string[StringType] | string[lib.t.Any]': ...
    @lib.t.overload  # type: ignore[no-overload-impl]
    def swapcase(self: lib.LiteralString) -> lib.LiteralString: ...
    @lib.t.overload
    def swapcase(self) -> 'string[StringType] | string[lib.t.Any]': ...  # type: ignore[misc]
    @lib.t.overload  # type: ignore[no-overload-impl]
    def title(self: lib.LiteralString) -> lib.LiteralString: ...
    @lib.t.overload
    def title(self) -> 'string[StringType] | string[lib.t.Any]': ...  # type: ignore[misc]
    def translate(  # type: ignore[empty-body]
        self,
        table: 'lib.builtins._TranslateTable',
        /
        ) -> 'string[StringType] | string[lib.t.Any]': ...
    @lib.t.overload  # type: ignore[no-overload-impl]
    def upper(self: lib.LiteralString) -> lib.LiteralString: ...
    @lib.t.overload
    def upper(self) -> 'string[StringType] | string[lib.t.Any]': ...  # type: ignore[misc]
    @lib.t.overload  # type: ignore[misc, no-overload-impl]
    def zfill(  # type: ignore[empty-body]
        self,
        width: lib.t.SupportsIndex,
        /
        ) -> 'string[StringType] | string[lib.t.Any]': ...  # type: ignore[misc]
    @lib.t.overload  # type: ignore[misc, no-overload-impl]
    def __add__(  # type: ignore[empty-body]
        self,
        value: str,
        /
        ) -> 'string[StringType] | string[lib.t.Any]': ...  # type: ignore[misc]
    def __getitem__(  # type: ignore[empty-body]
        self,
        key: lib.t.SupportsIndex | slice,
        /
        ) -> 'string[StringType] | string[lib.t.Any]': ...
    @lib.t.overload  # type: ignore[no-overload-impl]
    def __iter__(
        self: lib.LiteralString
        ) -> 'lib.t.Iterator[lib.LiteralString]': ...
    @lib.t.overload
    def __iter__(  # type: ignore[misc]
        self
        ) -> 'lib.t.Iterator[string[StringType] | string[lib.t.Any]]': ...
    @lib.t.overload  # type: ignore[no-overload-impl]
    def __mod__(
        self: lib.LiteralString,
        value: lib.t.Union[
            lib.LiteralString,
            tuple[lib.LiteralString, ...]
            ],
        /
        ) -> lib.LiteralString: ...
    @lib.t.overload
    def __mod__(
        self,
        value: lib.t.Any,
        /
        ) -> 'string[StringType] | string[lib.t.Any]': ...
    @lib.t.overload  # type: ignore[no-overload-impl]
    def __mul__(
        self: lib.LiteralString,
        value: lib.t.SupportsIndex,
        /
        ) -> lib.LiteralString: ...
    @lib.t.overload
    def __mul__(  # type: ignore[misc]
        self,
        value: lib.t.SupportsIndex,
        /
        ) -> 'string[StringType] | string[lib.t.Any]': ...
    @lib.t.overload  # type: ignore[no-overload-impl]
    def __rmul__(
        self: lib.LiteralString,
        value: lib.t.SupportsIndex,
        /
        ) -> lib.LiteralString: ...
    @lib.t.overload
    def __rmul__(  # type: ignore[misc]
        self,
        value: lib.t.SupportsIndex,
        /
        ) -> 'string[StringType] | string[lib.t.Any]': ...
    def __getnewargs__(  # type: ignore[empty-body]
        self
        ) -> 'tuple[string[StringType] | string[lib.t.Any]]': ...



"""Objects imports."""

from .. import core

__all__ = (
    'copy',
    'dataclass_transform',
    'inspect',
    *core.lib.__all__
    )

import copy
import inspect

from .. core . lib import *

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import dataclass_transform  # noqa  # type: ignore
else:  # pragma: no cover
    from typing import dataclass_transform



"""Typing types."""

__all__ = (
    'camelCase',
    'datetime',
    'numeric',
    'snake_case',
    'string',
    'AnyDict',
    'AnyField',
    'AnyOrForwardRef',
    'AnyOtherType',
    'AnyOtherTypeCo',
    'AnyString',
    'AnyType',
    'AnyTypeCo',
    'ArgsType',
    'Array',
    'ArrayType',
    'CamelDict',
    'Casing',
    'DataClassFields',
    'Enum',
    'FieldsTuple',
    'Immutable',
    'Literal',
    'Mapping',
    'MappingType',
    'NoneType',
    'NumberType',
    'Object',
    'ObjectType',
    'OptionalAnyDict',
    'OptionalGenericAlias',
    'PackageExceptionType',
    'Primitive',
    'Serial',
    'SnakeDict',
    'StrOrForwardRef',
    'StringFormat',
    'StringType',
    'Typed',
    'UnionGenericAlias',
    'VariadicArray',
    'VariadicArrayType',
    )

from . import lib
from . import obj

from . obj import (
    string,
    AnyType,
    AnyOtherType,
    AnyTypeCo,
    AnyOtherTypeCo,
    ArgsType,
    StringType
    )

if lib.t.TYPE_CHECKING:  # pragma: no cover
    from ... import objects  # noqa: F401
    from .. import exc  # noqa: F401

AnyOrForwardRef = lib.t.ForwardRef | lib.t.Any
StrOrForwardRef = lib.t.ForwardRef | str
OptionalGenericAlias = type(lib.t.Optional[str])
UnionGenericAlias = type(int | str)
Wrapper = obj.SupportsParams[lib.Unpack[ArgsType]]

camelCase = lib.t.NewType('camelCase', str)
snake_case = lib.t.NewType('snake_case', str)
datetime = lib.t.NewType('datetime', str)
numeric = lib.t.NewType('numeric', str)

AnyDict = dict[str, lib.t.Any]
AnyField: lib.t.TypeAlias = 'objects.Field[AnyType]'
AnyString = str | string[StringType]
Array = obj.ArrayProto[AnyType]
CamelDict: lib.t.TypeAlias = dict[string[camelCase], lib.t.Any]
Casing = camelCase | snake_case
DataClassFields: lib.t.TypeAlias = 'dict[string[snake_case], AnyField[lib.t.Any]]'  # noqa
Enum: lib.t.TypeAlias = 'Array[Immutable] | lib.enum.EnumMeta'
FieldsTuple: lib.t.TypeAlias = tuple[string[snake_case], ...]
Immutable: lib.t.TypeAlias = (
    bool
    | bytes
    | complex
    | lib.decimal.Decimal
    | lib.enum.Enum
    | float
    | lib.fractions.Fraction
    | int
    | lib.types.NoneType  # type: ignore[valid-type]
    | range
    | str
    | lib.enum.EnumMeta
    | frozenset['Immutable']
    | tuple['Immutable', ...]
    | lib.types.MappingProxyType['Immutable', 'Immutable']
    )
Literal = lib.t.Literal['*']
Mapping = obj.MappingProto[AnyType, AnyOtherType]
NoneType = lib.types.NoneType  # type: ignore[valid-type]
Object: lib.t.TypeAlias = 'objects.Object'
OptionalAnyDict = lib.t.Optional[dict[str, lib.t.Any]]
Primitive: lib.t.TypeAlias = bool | float | int | NoneType | str  # type: ignore[valid-type]
Serial: lib.t.TypeAlias = Primitive | dict[Primitive, 'Serial'] | list['Serial']  # noqa
SnakeDict: lib.t.TypeAlias = dict[string[snake_case], lib.t.Any]
StringFormat: lib.t.TypeAlias = camelCase | snake_case | datetime | numeric
Typed = obj.SupportsAnnotations
VariadicArray = obj.VariadicArrayProto[lib.Unpack[tuple[AnyType, ...]]]

ArrayType = lib.t.TypeVar('ArrayType', bound=Array)
MappingType = lib.t.TypeVar('MappingType', bound=Mapping)
NumberType = lib.t.TypeVar('NumberType', bound=lib.numbers.Number)
ObjectType = lib.t.TypeVar('ObjectType', bound=Object)
PackageExceptionType = lib.t.TypeVar(
    'PackageExceptionType',
    bound='exc.BasePackageException',
    covariant=True,
    )
VariadicArrayType = lib.t.TypeVar('VariadicArrayType', bound=VariadicArray)



"""Docs imports."""

from .. import core

__all__ = (
    'configparser',
    'shutil',
    'string',
    *core.lib.__all__
    )

import configparser
import shutil
import string

from .. core . lib import *



"""Object module."""

__all__ = (
    'Object',
    )

from ... import core

from .. import cfg
from .. import exc
from .. import lib
from .. import metas
from .. import typ
from .. import utl

if lib.t.TYPE_CHECKING:  # pragma: no cover
    from .. import queries


class Constants(cfg.Constants):
    """Constant values specific to this file."""


@lib.dataclass_transform(
    field_specifiers=(typ.Field, )
    )
class ObjectBase(metaclass=metas.Meta):
    """Base for Objects, used for typing purposes."""

    __annotations__: typ.SnakeDict
    __dict__: dict[typ.AnyString, lib.t.Any]
    __dataclass_fields__: lib.t.ClassVar[typ.DataClassFields]
    __heritage__: lib.t.ClassVar[tuple['metas.Meta', ...]]

    enumerations: lib.t.ClassVar[dict[str, tuple[typ.Primitive, ...]]]
    fields: lib.t.ClassVar[typ.FieldsTuple]
    hash_fields: lib.t.ClassVar[typ.FieldsTuple]

    def __repr__(self) -> str:
        """
        Return constructor represented as a neatly formatted JSON string.

        """

        return core.codecs.utl.serialize(self)

    def __init__(
        self,
        class_as_dict: lib.t.Optional[dict[typ.string[typ.snake_case], lib.t.Any]] = None,
        /,
        **kwargs: lib.t.Any
        ):
        ckwargs = {
            cname: value
            for name, value
            in kwargs.items()
            if (cname := core.strings.utl.cname_for(name, self.fields))
            }

        if isinstance(class_as_dict, lib.t.Mapping):
            class_as_cdict = {
                cname: value
                for name, value
                in class_as_dict.items()
                if (cname := core.strings.utl.cname_for(name, self.fields))
                }
            ckwargs |= class_as_cdict

        for cname, field in self.__dataclass_fields__.items():
            if cname not in ckwargs:
                ckwargs[cname] = field.factory()

        for cname, value in ckwargs.items():
            setattr(self, cname, value)

        self.__post_init__()

    def __post_init__(self) -> None:
        """Method that will always run after instantiation."""

    def __delitem__(self, __key: lib.t.Any) -> lib.t.Optional[lib.Never]:
        """Reset current value for key to field default."""

        if (
            isinstance(__key, str)
            and (k := core.strings.utl.cname_for(__key, self.fields))
            ):
            self.pop(k, None)
            return None
        else:
            raise KeyError(__key)

    def __getitem__(self, __key: lib.t.Any, /) -> lib.t.Any:
        """Return field value dict style."""

        if (
            isinstance(__key, str)
            and (k := core.strings.utl.cname_for(__key, self.fields))
            ):
            value = getattr(
                self,
                k,
                (
                    field_.factory()
                    if typ.utl.check.is_field(
                        field_ := self.__dataclass_fields__[k]
                        )
                    else field_['default']
                    )
                )
            if (
                isinstance(value, Object)
                and (
                    callers := lib.t.cast(
                        lib.types.FrameType,
                        lib.t.cast(
                            lib.types.FrameType,
                            lib.inspect.currentframe()
                            ).f_back
                        ).f_code.co_names
                    )
                and 'dict' in callers
                and (
                    callers[0] == 'dict'
                    or (
                        callers[callers.index('dict') - 1]
                        != 'to_dict'
                        )
                    )
                ):
                return value.to_dict()
            elif (
                typ.utl.check.is_array(value)
                and (
                    callers := lib.t.cast(
                        lib.types.FrameType,
                        lib.t.cast(
                            lib.types.FrameType,
                            lib.inspect.currentframe()
                            ).f_back
                        ).f_code.co_names
                    )
                and 'dict' in callers
                and (
                    callers[0] == 'dict'
                    or (
                        callers[callers.index('dict') - 1]
                        != 'to_dict'
                        )
                    )
                ):
                return value.__class__(
                    item.to_dict()
                    if typ.utl.check.is_object(item)
                    else item
                    for item
                    in value
                    )
            elif (
                typ.utl.check.is_mapping(value)
                and (
                    callers := lib.t.cast(
                        lib.types.FrameType,
                        lib.t.cast(
                            lib.types.FrameType,
                            lib.inspect.currentframe()
                            ).f_back
                        ).f_code.co_names
                    )
                and 'dict' in callers
                and (
                    callers[0] == 'dict'
                    or (
                        callers[callers.index('dict') - 1]
                        != 'to_dict'
                        )
                    )
                ):
                return value.__class__(
                    {
                        (
                            k.to_dict()
                            if typ.utl.check.is_object(k)
                            else k
                            ): (
                                v.to_dict()
                                if typ.utl.check.is_object(v)
                                else v
                                )
                        for k, v
                        in value.items()
                        }
                    )
            else:
                return value
        else:
            raise KeyError(__key)

    def __setitem__(
        self,
        __key: str,
        __value: lib.t.Any
        ) -> lib.t.Optional[lib.Never]:
        """Set field value dict style."""

        if (k := core.strings.utl.cname_for(__key, self.fields)):
            setattr(self, k, __value)
            return None
        else:
            raise KeyError(__key)

    def __contains__(self, __key: lib.t.Any, /) -> bool:
        """Return `True` if `__key` is a field for self."""

        return bool(core.strings.utl.cname_for(__key, self.fields))

    def __len__(self) -> int:
        """Return count of fields."""

        return len(self.fields)

    def __hash__(self) -> int:
        return hash(
            Constants.DELIM.join(
                [
                    '.'.join((k, str(v)))
                    for k
                    in self.hash_fields
                    if (v := self.get(k))
                    ]
                )
            )

    def __bool__(self) -> bool:
        """Determine truthiness by diff with default field values."""

        return bool(self - self.__class__())

    @lib.t.overload
    def __eq__(
        self,
        other: 'typ.AnyField[lib.t.Any]'
        ) -> bool: ...
    @lib.t.overload
    def __eq__(
        self: object,
        other: object
        ) -> bool: ...
    @lib.t.overload
    def __eq__(
        self,
        other: lib.t.Any
        ) -> lib.t.Union[bool, 'queries.EqQueryCondition', lib.Never]: ...
    def __eq__(
        self,
        other: lib.t.Union[object, lib.t.Any]
        ) -> lib.t.Union[bool, 'queries.EqQueryCondition', lib.Never]:

        return hash(self) == hash(other)

    @lib.t.overload
    def __ne__(
        self,
        other: 'typ.AnyField[lib.t.Any]'
        ) -> bool: ...
    @lib.t.overload
    def __ne__(
        self: object,
        other: object
        ) -> bool: ...
    @lib.t.overload
    def __ne__(
        self,
        other: lib.t.Any
        ) -> lib.t.Union[bool, 'queries.NeQueryCondition', lib.Never]: ...
    def __ne__(
        self,
        other: lib.t.Union[object, lib.t.Any, 'typ.AnyField[lib.t.Any]']
        ) -> lib.t.Union[bool, 'queries.NeQueryCondition', lib.Never]:

        return hash(self) != hash(other)

    def __sub__(self, other: lib.Self) -> typ.SnakeDict:
        """Calculate diff between same object types."""

        diff: typ.SnakeDict = {}
        for field in self.fields:
            if self[field] != other[field]:
                diff[field] = other[field]

        return diff

    def __iter__(
        self
        ) -> lib.t.Iterator[tuple[typ.string[typ.snake_case], lib.t.Any]]:
        """
        Return an iterator of keys and values like a `dict`.

        ---

        Removes any suffixed underscores from field names (`_`).

        """

        for k, v in self.items():
            yield k, v

    @lib.t.overload
    def __lshift__(
        self,
        other: typ.obj.ObjectLike,
        ) -> lib.Self: ...
    @lib.t.overload
    def __lshift__(
        self,
        other: lib.t.Any,
        ) -> lib.t.Union[
            'queries.ContainsQueryCondition',
            lib.Self,
            lib.Never
            ]: ...
    def __lshift__(
        self,
        other: typ.obj.ObjectLike | lib.t.Any
        ) -> lib.t.Union[
            lib.Self,
            'queries.ContainsQueryCondition',
            lib.Never
            ]:
        """
        Interpolate values from other if populated with non-default \
        and return a new instance without mutating self or other.

        """

        if not all(field in other for field in self.__dataclass_fields__):
            raise exc.InvalidObjectComparisonError(self, other)
        else:
            object_ = self.__class__()
            for field, __field in self.__dataclass_fields__.items():
                if typ.utl.check.is_field(self):
                    default_value = __field['default']
                else:
                    default_value = __field.factory()
                if (
                    self[field] == default_value
                    and other[field] != default_value
                    ):
                    object_[field] = other[field]
                else:
                    object_[field] = self[field]
            return object_

    def __rshift__(
        self,
        other: typ.obj.ObjectLike
        ) -> lib.Self:
        """
        Overwrite values from other if populated with non-default \
        and return a new instance without mutating self or other.

        """

        object_ = self.__class__()
        for field, __field in self.__dataclass_fields__.items():
            if other[field] != __field.factory():
                object_[field] = other[field]
            else:
                object_[field] = self[field]
        return object_

    def __reversed__(self) -> lib.t.Iterator[typ.string[typ.snake_case]]:
        """
        Return a reversed iterator of keys like a `dict`.

        ---

        Removes any suffixed underscores from field names (`_`).
        
        """

        for field in reversed(sorted(self.keys())):
            yield field

    def __copy__(self) -> lib.Self:
        """Return a copy of the instance."""

        return self.__class__(dict(self))

    def __deepcopy__(
        self,
        memo: lib.t.Optional[typ.AnyDict] = None
        ) -> lib.Self:
        """Return a deep copy of the instance."""

        return self.__copy__()

    def __getstate__(self) -> typ.SnakeDict:
        return dict(self)

    def __setstate__(self, state: typ.SnakeDict) -> None:
        other: typ.obj.ObjectLike = self.__class__(state)
        self.update(other)
        return None

    def __ior__(self, other: typ.obj.ObjectLike, /) -> lib.Self:
        self.update(other)
        return self

    def get(
        self,
        __key: typ.AnyString,
        __default: typ.AnyType = None
        ) -> lib.t.Any | typ.AnyType:
        """Return value by key if exists, otherwise default."""

        if (k := core.strings.utl.cname_for(__key, self.fields)):
            return self[k]
        else:
            return __default

    def copy(self) -> lib.Self:
        """Return a copy of the instance."""

        return self.__copy__()

    @classmethod
    def fromkeys(
        cls,
        __keys: lib.t.Iterable[typ.string[typ.snake_case]],
        /
        ) -> lib.Self:
        """
        Return an object instance from keys like a `dict`.

        ---

        Removes any suffixed underscores from field names (`_`).

        """

        return cls()

    @classmethod
    def keys(cls) -> lib.t.KeysView[typ.string[typ.snake_case]]:
        """
        Return an iterator of keys like a `dict`.

        ---

        Removes any suffixed underscores from field names (`_`).

        """

        k_: typ.string[typ.snake_case]
        return lib.t.KeysView(
            {
                k_: v
                for k, v
                in cls.__dataclass_fields__.items()
                if (k_ := k.rstrip('_'))
                }
            )

    def items(self) -> lib.t.ItemsView[typ.string[typ.snake_case], lib.t.Any]:
        """
        Return an iterator of keys and values like a `dict`.

        ---

        Removes any suffixed underscores from field names (`_`).

        """

        return self.to_dict().items()

    def pop(
        self,
        __key: str,
        /,
        __default: typ.AnyType = Constants.UNDEFINED
        ) -> typ.AnyType | lib.t.Any | lib.Never:
        """Return current value for key and reset instance value to field default."""

        if (cname := core.strings.utl.cname_for(__key, self.fields)):
            value = self[cname]
            self[cname] = self.__dataclass_fields__[cname].factory()
            return value
        elif __default == Constants.UNDEFINED:
            raise KeyError
        else:
            return __default

    def setdefault(
        self,
        __key: str,
        __value: lib.t.Any
        ) -> lib.t.Optional[lib.Never]:
        """Set value for key if unset; otherwise do nothing."""

        if (
            (k := core.strings.utl.cname_for(__key, self.fields))
            and (
                (
                    _value := self.get(k, Constants.UNDEFINED)
                    ) == Constants.UNDEFINED
                or _value == self.__dataclass_fields__[k].factory()
                )
            ):
            self[k] = __value
        elif not k:
            raise KeyError

        return None

    def update(self, other: typ.obj.ObjectLike, /) -> None:
        """Update values like a `dict`."""

        for k, v in other.items():
            if core.strings.utl.cname_for(k, self.fields):
                self[k] = v

        return None

    def values(self) -> lib.t.ValuesView[lib.t.Any]:
        """Return an iterator of values like a `dict`."""

        return lib.t.ValuesView(dict(self))

    @lib.t.overload
    def to_dict(
        self,
        camel_case: lib.t.Literal[False] = False,
        include_null: bool = True
        ) -> typ.SnakeDict: ...
    @lib.t.overload
    def to_dict(
        self,
        camel_case: lib.t.Literal[True],
        include_null: bool
        ) -> typ.CamelDict: ...
    @lib.t.overload
    def to_dict(
        self,
        camel_case: bool,
        include_null: bool
        ) -> 'typ.SnakeDict | typ.CamelDict': ...
    def to_dict(
        self,
        camel_case: bool = False,
        include_null: bool = True
        ) -> 'typ.SnakeDict | typ.CamelDict':
        """
        Same as `dict(Object)`, but gives fine-grained control over \
        casing and inclusion of `null` values.

        ---

        If specified, keys may optionally be converted to camelCase.

        `None` values may optionally be discarded as well.

        ---

        Removes any suffixed underscores from field names (`_`) and \
        recursively pops any key, value pairs prefixed with single \
        underscores (`_`).

        """

        d = {
            k: v
            for k
            in self.fields
            if (v := self[k]) is not None
            or (include_null and v is None)
            }
        as_dict: typ.SnakeDict = {}
        for key, value in d.items():
            if isinstance(value, Object):
                as_dict[key] = value.to_dict(camel_case, include_null)
            elif typ.utl.check.is_array(value):
                as_dict[key] = value.__class__(
                    (
                        v.to_dict(camel_case, include_null)
                        if isinstance(v, Object)
                        else v
                        for v
                        in value
                        if v is not None
                        or include_null
                        )
                    )
            elif typ.utl.check.is_mapping(value):
                as_dict[key] = value.__class__(
                    **{
                        (
                            core.strings.utl.snake_case_to_camel_case(k.strip('_'))
                            if (camel_case and isinstance(k, str))
                            else k
                            ): (
                                v.to_dict(camel_case, include_null)
                                if isinstance(v, Object)
                                else v
                                )
                        for k, v
                        in value.items()
                        if utl.is_public_field(k)
                        and v is not None
                        or include_null
                        }
                    )
            else:
                as_dict[key] = value

        if camel_case:
            return {
                core.strings.utl.snake_case_to_camel_case(k.strip('_')): v
                for k, v
                in as_dict.items()
                }
        else:
            k_: typ.string[typ.snake_case]
            snake_dict: typ.SnakeDict = {
                k_: v
                for k, v
                in as_dict.items()
                if (k_ := k.rstrip('_'))
                }
            return snake_dict


@lib.dataclass_transform(
    kw_only_default=True,
    field_specifiers=(typ.Field, )
    )
class Object(ObjectBase):
    """
    Base Object.

    ---

    Usage
    -----

    * Subclass to create objects for your application.

    General Recommendations
    -----------------------

    * Ideally, objects should be 1:1 with their counterparts in the \
    data store from which they are originally sourced (even \
    if that data store is your own database, and even if that \
    data is not ostensibly stored in a 1:1 manner, as is the case \
    with most relational databases).
        \
        * For example, if there is a SQL table called `pets` \
        with the schema below, you would want to create \
        a corresponding `python representation` similar to \
        the following.

    #### pets table

    ```
    | id  | name     | type   |
    | --- | -------- | ------ |
    | a1  | fido     | dog    |
    | a2  | garfield | cat    |
    | a3  | sophie   | dog    |
    | a4  | stripes  | turtle |

    ```

    #### python representation

    ```py
    import fqr


    class Pet(fqr.Object):
        \"""A pet.\"""

        id_: fqr.Field[str] # Trailing underscores are special
                             # in fqr, check the documentation
                             # below for more detail.
        name: fqr.Field[str] = 'Fido'  # Setting = 'Fido' will mean that
                                       # all Pet() instances will be
                                       # named 'Fido' by default.
        type: fqr.Field[str]  # You can make a field 'required' by
                              # not specifying a default value.

    ```

    ---

    Special Rules
    -------------

    #### Default Values
    Subclassed (derivative) objects should include default values for \
    all fields specified. In cases where a default value is not specified, \
    `None` will be used instead and the field will be assumed to be \
    'required' for all downstream purposes (ex. as a query parameter \
    for HTTP requests) unless otherwise specified explicitly.

    #### Type Annotations
    Type annotations are required and must be a generic `Field[type]`. \
    For example: `Field[int]`, `Field[str]`, `Field[str | bool]`.

    * Not only is this best practice, these are leveraged downstream \
    to do things like auto-document and auto-generate API's.

    #### Uniform Casing
    ALL Fields must be either camelCase or snake_case, with the only \
    exception being that fields may begin with an underscore '_', so long \
    as all following characters adhere to camelCase or snake_case conventions.

    #### Underscore Prefix for Private Fields
    Fields that begin with an underscore '_' will be ignored on \
    conversion to / from DBO, REST, and JSON representations, \
    unless the field ends with 'id' or 'id_' (case insensitive), \
    in which case it will still be converted.

    * This follows the broader pattern of flagging methods and \
    attributes as private / internal to a system with a preceding \
    underscore. It should be expected that end users of your \
    system will not need to interact with these fields.

    #### Underscore Suffix for Reserved Keyword Fields
    Fields with a trailing underscore '_' will automatically have \
    the trailing underscore removed on conversion to / from \
    DBO, REST, and JSON representations.

    * This allows for python keywords, such as `in_`, to be used \
    as object fields, where they would otherwise raise errors \
    without the proceeding underscore.

    * On translation to and from dictionaries, keys without \
    underscores will still be checked against these fields -- \
    so, a dictionary with key `in` will correctly map to the `in_` \
    field on the Object. See below for more detail.

    ```py
    import fqr


    class Pet(fqr.Object):
        \"""A pet.\"""

        id_: fqr.Field[str]
        _alternate_id: fqr.Field[str]

        name: fqr.Field[str]
        type: fqr.Field[str]
        in_: fqr.Field[str]
        is_tail_wagging: fqr.Field[bool] = True


    # This means each of the below will work.
    bob_the_dog = Pet(
        id='abc123',
        _alternate_id='dog1',
        name='Bob',
        type='dog',
        in_='timeout',
        is_tail_wagging=False
        )
    bob_the_dog = Pet(
        {
            'id': 'abc123',
            '_alternate_id': 'dog1',
            'name': 'Bob',
            'type': 'dog',
            'in': 'timeout',
            'is_tail_wagging': False
            }
        )

    # And so would this, since translation
    # automatically handles camelCase to
    # snake_case conversions.
    bob_the_dog = Pet(
        {
            'id': 'abc123',
            'alternateId': 'dog1',
            'name': 'Bob',
            'type': 'dog',
            'in': 'timeout',
            'isTailWagging': False
            }
        )

    ```

    ---

    Special Method Usage
    --------------------

    Objects have been designed to be almost interchangable with \
    dictionaries. The primary difference is that values cannot be \
    assigned to keys unless you define them on the Object's class \
    definition itself.
    * This is done to automatically maximize the efficiency of your \
    application's memory footprint. Feel free to read more about \
    python [slots](https://wiki.python.org/moin/UsingSlots) to better \
    understand why this is necessary.

    ```py
    import fqr


    class Pet(fqr.Object):  # noqa

        name: fqr.Field[str]


    dog = Pet(name='Fido')

    # The below would return the string, 'Fido'.
    dog['name']

    # The following would set the dog's name to something else.
    dog.setdefault('name', 'Arnold')
    assert dog.name == 'Arnold'
    dog.setdefault('name', 'Buddy')
    assert dog.name == 'Arnold'
    dog['name'] = 'Buddy'
    assert dog.name == 'Buddy'
    assert dog['name'] == 'Buddy'

    # The following all work exactly the same as with a dictionary.
    # (in the below, key will be 'name' and value 'Fido').
    for key, value in dog.items():
        break

    for key in dog.keys():
        break

    for value in dog.values():
        break

    # But the following will raise a KeyError.
    dog['field_that_does_not_exist'] = 'Buddy'

    # And so would this, since fields can only be added
    # or removed on the class definition of Pet itself.
    dog.setdefault('field_that_does_not_exist', 'Buddy')

    ```

    Object truthiness will evaluate to True if any values for \
    the Object instance are different from default values, \
    otherwise False.

    ```py
    if Object:
    ```

    Objects are designed to display themselves as neatly \
    formatted JSON on calls to `__repr__`.

    ```py
    print(Object)
    ```

    Updates Object1 with values from Object2 if they \
    are a non-default value for the object.

    ```py
    Object1 << Object2
    ```

    Overwrites Object1 values with those from Object2 \
    if they are a non-default value for the object.

    ```py
    Object1 >> Object2
    ```

    Returns a dictionary with {fieldName: fieldValue2} for \
    any fields that differ between the two Objects.

    ```py
    Object1 - Object2
    ```

    Get value for Object field.

    ```py
    value = Object['field']
    ```

    Set value for Object field.

    ```py
    Object['field'] = value
    ```

    Returns True if any one of field, _field, field_, or _field_ \
    is a valid field for the Object, otherwise False.

    ```py
    field in Object
    ```

    Same as `len(Object.fields)`.

    ```py
    len(Object)
    ```

    """

    class_as_dict: lib.t.Final[
        lib.t.Optional[dict[typ.string[typ.snake_case], lib.t.Any]]
        ] = None
    """
    Instantiate class directly from passed `dict` (assumed to be \
    version of class in `dict` form).

    """



"""Codecs enumerations."""

from .. import enm

__all__ = (
    'ParseErrorRef',
    *enm.__all__
    )

from .. enm import *

from . import lib
from . import typ


class ParseErrorRef(lib.enum.Enum):
    """
    Valid ErrorRef Enumeration.

    """

    bool_decode           = typ.ErrorRef(
        'Could not decode valid JSON string to python `bool`.'
        )
    invalid_arr_decode    = typ.ErrorRef(
        'Could not decode valid JSON array to python: `invalid item type(s)`.'
        )
    invalid_arr_len       = typ.ErrorRef(
        'Could not decode valid JSON array to python: `invalid array length`.'
        )
    invalid_json          = typ.ErrorRef(
        'Could not deserialize string as valid JSON.'
        )
    invalid_keys_decode   = typ.ErrorRef(
        'Could not decode valid JSON object to python: `invalid keys`.'
        )
    invalid_map_decode    = typ.ErrorRef(
        ' '.join(
            (
                'Could not decode valid JSON object to python:',
                '`invalid item type(s)`.'
                )
            )
        )
    invalid_values_decode = typ.ErrorRef(
        'Could not decode valid JSON object to python: `invalid values`.'
        )
    number_decode         = typ.ErrorRef(
        ' '.join(
            (
                'Could not decode valid JSON string to python',
                '`numbers.Number`.'
                )
            )
        )
    null_decode           = typ.ErrorRef(
        'Could not decode valid JSON string to python `None`.'
        )
    datetime_decode       = typ.ErrorRef(
        'Could not decode value to valid `datetime`.'
        )
    literal_decode        = typ.ErrorRef(
        'Could not decode value to valid `Literal`.'
        )
    value_decode          = typ.ErrorRef('Could not decode invalid value.')



"""Strings constants."""

__all__ = (
    'Constants',
    )

from .. import cfg

from . import lib


class Constants(cfg.Constants):
    """Constant values shared across core strings modules."""

    MAX_CHARS = int(lib.os.getenv('MAX_CHARS', 384))
    """
    Default, package-wide maximum string length [before it is wrapped \
    into a multi-line string].

    ---

    Used for `__repr__`, log messages, etc.

    """

    CUTOFF_LEN = int(lib.os.getenv('CUTOFF_LEN', 12))
    """
    Default, package-wide maximum multi-line string line count [before \
    `list[str]` is trimmed and an ellipsis (...) is appended as its \
    final item].

    ---

    Used for `__repr__`, log messages, etc.

    """

    WRAP_WIDTH = int(lib.os.getenv('WRAP_WIDTH', 64))
    """
    Default, package-wide maximum multi-line string line count [before \
    `list[str]` is trimmed and an ellipsis (...) is appended as its \
    final item].

    ---

    Used for `__repr__`, log messages, etc.

    """

    M_LINE_TOKEN = '[[MULTI_LINE_STRING_AS_ARRAY]]'
    """
    Token pre-pended to a wrapped, multi-line string to indicate that \
    the value was originally a string.

    """



"""
Overview
========

**Author:** dan@1howardcapital.com

**Summary:** Zero-dependency python framework for object oriented development.

---

Usage
-----

```py
import fqr

```

"""

__all__ = (
    'cfg',
    'codecs',
    'enm',
    'exc',
    'lib',
    'strings',
    'typ',
    )

from . import cfg
from . import codecs
from . import enm
from . import exc
from . import lib
from . import strings
from . import typ



"""Modules for handling strings."""

__all__ = (
    'cfg',
    'enm',
    'exc',
    'lib',
    'obj',
    'typ',
    'utl',
    )

from . import cfg
from . import enm
from . import exc
from . import lib
from . import obj
from . import typ
from . import utl



"""Core typing modules."""

__all__ = (
    'camelCase',
    'datetime',
    'numeric',
    'obj',
    'snake_case',
    'string',
    'utl',
    'AnyDict',
    'AnyField',
    'AnyOrForwardRef',
    'AnyOtherType',
    'AnyOtherTypeCo',
    'AnyString',
    'AnyType',
    'AnyTypeCo',
    'ArgsType',
    'Array',
    'ArrayType',
    'CamelDict',
    'Casing',
    'DataClassFields',
    'Enum',
    'FieldsTuple',
    'Immutable',
    'Literal',
    'Mapping',
    'MappingType',
    'NoneType',
    'NumberType',
    'Object',
    'ObjectType',
    'OptionalAnyDict',
    'PackageExceptionType',
    'Primitive',
    'Serial',
    'SnakeDict',
    'StrOrForwardRef',
    'StringFormat',
    'StringType',
    'Typed',
    'UnionGenericAlias',
    'VariadicArray',
    'VariadicArrayType',
    )

from . import obj
from . import utl

from . typ import *



"""Docs utility functions."""

__all__ = (
    'document',
    )

from .. import cli

from . import cfg
from . import enm
from . import lib
from . import obj
from . import static


class Constants(cfg.Constants):
    """Documentation utils specific constants."""


def _do_api_doc(
    package_name: str,
    output_dir: str,
    is_namespace_package: bool,
    include_private_modules: bool,
    make_index: bool,
    ) -> None:

    if lib.os.path.exists(lib.os.path.join(output_dir, 'docs')):
        lib.shutil.rmtree(lib.os.path.join(output_dir, 'docs'))
    lib.os.makedirs(lib.os.path.join(output_dir, 'docs', 'source', '_static'))

    flags = [
        '-d', '4',  # noqa
        '-e',
        '-f',
        '-M',
        ]
    if make_index:
        flags.append('--tocfile')
        flags.append('index')
    else:
        flags.append('-T')
    if is_namespace_package:
        flags.append('--implicit-namespaces')
    if include_private_modules:
        flags.append('-P')

    import sphinx.ext.apidoc

    sphinx.ext.apidoc.main(
        (
            *flags,
            '-o',
            lib.os.path.join(output_dir, 'docs', 'source'),
            package_name,
            )
        )

    return None


def _export_conf(
    package_name: str,
    package_version: str,
    author: str,
    sphinx_theme: str,
    add_module_names: bool,
    autodoc_inherit_docstrings: bool,
    output_dir: str,
    ) -> None:

    template = lib.string.Template(static.config_template)
    rendered = template.safe_substitute(
        {
            'add_module_names': add_module_names,
            'author_name': author,
            'autodoc_inherit_docstrings': autodoc_inherit_docstrings,
            'package_name': package_name,
            'package_version': package_version,
            'sphinx_theme': sphinx_theme,
            }
        )

    out_file = lib.os.path.join(output_dir, 'docs', 'source', 'conf.py')
    with open(out_file, 'w') as f:
        f.write(rendered)

    return None


def _do_build(
    output_dir: str,
    ) -> None:

    import sphinx.cmd.build

    sphinx.cmd.build.main(
        (
            '-a',
            '-E',
            lib.os.path.join(output_dir, 'docs', 'source'),
            lib.os.path.join(output_dir, 'docs', 'html')
            )
        )

    return None


def document(
    package: str,
    version: str,
    author: str,
    output_dir: str,
    favicon_path: str,
    logo_path: str,
    sphinx_theme: str,
    is_namespace_package: bool,
    add_module_names: bool,
    autodoc_inherit_docstrings: bool,
    no_cleanup: bool,
    include_private_modules: bool,
    make_index: bool,
    readme_path: lib.t.Optional[str],
    no_include_meta_tags: bool,
    no_include_robots: bool,
    site_map_urls: list[str],
    site_map_change_freq: str,
    ) -> None:
    """
    CLI entrypoint for documenting a python package in wiki style.

    ---

    Requirements
    ------------

    Leverages and requires the sphinx library (and an extension to
    handle markdown), which is technically a third-party dependency, \
    even if sphinx is used to [document python itself](https://devguide.python.org/documentation/start-documenting/#building-doc).

    They can be installed using the following command.

    `$ pip install fqr[docs]`

    ---

    Usage
    -----

    Specify the name of the package (or [relative] path to the package) \
    as the first positional argument.

    For example, if we are in the root level directory of a standard \
    python repo called `package`, then the following command will...

    `$ fqr docs .`

    ...create a `/docs` sub-directory in the directory from which the \
    script is executed, with all documentation for `package` fully \
    generated within.

    For additional command options and argument flags, please use:

    `$ fqr docs --help`

    """

    try:
        import commonmark  # type: ignore
        import sphinx  # type: ignore
    except (ImportError, ModuleNotFoundError):
        cli.obj.root_parser.exit(
            1,
            '\n'.join(
                (
                    'Missing a required third-party dependency.',
                    'Install with `$ pip install fqr[docs]`.',
                    'Exiting...'
                    )
                )
            )

    match = obj.Pattern.PrefixPattern.match(package)
    is_local_package = bool(match)

    package_name = obj.Pattern.PrefixPattern.sub('', package)

    if is_local_package:

        for _, dirnames, filenames in lib.os.walk('.'):
            is_pyproject_cfg = 'pyproject.toml' in filenames
            is_src_package = 'src' in dirnames
            break

        if is_pyproject_cfg and not package_name:
            parser = lib.configparser.ConfigParser()
            parser.read('pyproject.toml')
            package_name = parser['project']['name'][1:-1]
        elif not package_name:
            with open('setup.py', 'r') as f:
                in_setup_fn = False
                for line in f.readlines():
                    if 'setup(' in line:
                        in_setup_fn = True
                    if in_setup_fn and 'name' in line:
                        _, _, partition = line.partition('name=')
                        re_match_name = obj.Pattern.MinOneWordPattern.match(
                            partition
                            )
                        if re_match_name is not None:
                            package_name = re_match_name.group()
                            break

    else:
        is_src_package = False

    package_name = obj.Pattern.PathPattern.sub('.', package_name)
    re_match_root = obj.Pattern.MinOneWordPattern.match(package_name)
    if re_match_root is not None:
        package_root = re_match_root.group()
    else:
        package_root = package_name

    _do_api_doc(
        (
            lib.os.path.join('src', package_root)
            if is_src_package
            else package_root
            ),
        output_dir,
        is_namespace_package,
        include_private_modules,
        make_index or bool(readme_path),
        )

    source_dir = lib.os.path.join(output_dir, 'docs', 'source')
    static_dir = lib.os.path.join(source_dir, '_static')
    lib.shutil.copyfile(
        favicon_path,
        lib.os.path.join(static_dir, 'favicon.ico'),
        )
    lib.shutil.copyfile(
        logo_path,
        lib.os.path.join(static_dir, 'logo.png'),
        )

    # Below is done to fix static asset rendering on github pages.
    # See: https://github.com/sphinx-doc/sphinx/issues/2202
    with open(
        lib.os.path.join(output_dir, 'docs', '.nojekyll'),
        'w'
        ) as jkl_file:
        jkl_file.write('')

    # Remove generated files irrelevant to the
    # targeted namespace.
    if is_namespace_package:
        for pre_file_name in lib.os.listdir(source_dir):
            if (
                pre_file_name.endswith('.rst')
                and not pre_file_name.endswith('index.rst')
                and not pre_file_name.startswith(package_name)
                ):
                lib.os.remove(
                    lib.os.path.join(
                        output_dir,
                        'docs',
                        'source',
                        pre_file_name
                        )
                    )

    _export_conf(
        package_name,
        version,
        author,
        sphinx_theme,
        add_module_names,
        autodoc_inherit_docstrings,
        output_dir,
        )

    # Append meta tags to files.
    if not no_include_meta_tags:
        for file_name in lib.os.listdir(source_dir):
            if not file_name.endswith('.rst'):
                continue
            module_name = file_name.removesuffix('.rst')
            if (
                (
                    module_name == package_root
                    and not (make_index or readme_path)
                    )
                or file_name.startswith('index')
                ):
                txt = ''.join(
                    (
                        '      ',
                        f"Documentation homepage for the '{package_name}' ",
                        'project.'
                        )
                    )
            else:
                txt = ' '.join(
                    (
                        f"      Documentation page for the '{module_name}'",
                        f"component of the '{package_name}' project."
                        )
                    )
            lines = [
                '',
                '.. meta::',
                '   :description lang=en:',
                txt,
                '',
                ]
            with open(
                lib.os.path.join(source_dir, file_name),
                'a'
                ) as file_buf:
                file_buf.write('\n'.join(lines))

    _do_build(output_dir)

    # Optionally include sitemap.
    if site_map_urls:
        html_dir = lib.os.path.join(output_dir, 'docs', 'html')
        pages: list[str] = [
            html_file_name
            for html_file_name
            in lib.os.listdir(html_dir)
            if (
                html_file_name.endswith('.html')
                and (
                    html_file_name.startswith(package_root)
                    or html_file_name.startswith('index')
                    )
                )
            ]
        page_depths = [
            len(page_name.split('.'))
            for page_name
            in pages
            ]
        min_depth = min(page_depths)
        max_depth = max(page_depths)
        lastmod = lib.datetime.datetime.now(
            lib.datetime.timezone.utc
            ).isoformat()

        url_set: list[str] = [
            (
                '<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9"'
                ' xmlns:xhtml="http://www.w3.org/1999/xhtml">'
                ),
            ]
        for site_map_url in site_map_urls:
            parsed = lib.urllib.parse.urlparse(site_map_url)
            for page in pages:
                page_depth = len(page.split('.'))
                page_loc = '/'.join((site_map_url.rstrip('/'), page))
                priority = (max_depth - page_depth + min_depth) / max_depth
                xml_url = '\n'.join(
                    (
                        '<url>',
                        f'    <loc>{page_loc}</loc>',
                        f'    <changefreq>{site_map_change_freq}</changefreq>',
                        f'    <lastmod>{lastmod}</lastmod>',
                        f'    <priority>{priority!s}</priority>',
                        '</url>',
                        )
                    )
                url_set.append(xml_url)

        xml_path = lib.os.path.join(output_dir, 'docs', 'html', 'sitemap.xml')
        with open(xml_path, 'w') as xml_file:
            xml_file.write(
                '\n'.join(
                    (
                        '<?xml version="1.0" encoding="UTF-8"?>',
                        *url_set,
                        '</urlset>'
                        )
                    )
                )

    # Include robotos.txt unless otherwise specified.
    if not no_include_robots:
        robots_path = lib.os.path.join(
            output_dir,
            'docs',
            'html',
            'robots.txt'
            )
        with open(robots_path, 'w') as robots_file:
            robots_txt = '\n'.join(
                (
                    'User-agent: *',
                    'Allow: /',
                    *enm.NoCrawlPath._value2member_map_,
                    )
                )
            if site_map_urls:
                parsed = lib.urllib.parse.urlparse(site_map_urls[0])
                url_root = '://'.join((parsed.scheme, parsed.hostname or ''))
                path = parsed.path.strip('/')
                robots_txt += f'\n\nSitemap: {url_root}/{path}/sitemap.xml'
            robots_file.write(robots_txt)

    # Inject readme.md if exists.
    if readme_path:
        with open(readme_path, 'r', errors='ignore') as md_file:
            md = md_file.read()

        html = commonmark.commonmark(md, format='html')

        index_path = lib.os.path.join(
            output_dir,
            'docs',
            'html',
            'index.html'
            )
        with open(index_path, 'r') as html_file:
            index_html = html_file.read()

        new_lines = []
        for line in index_html.split('\n'):
            if line.strip() == f'<section id="{package_name}">':
                new_lines.extend(html.split('\n'))
                line = f'<section id="{package_name}" hidden=true>'
            new_lines.append(line)

        with open(index_path, 'w') as html_file:
            html_file.write('\n'.join(new_lines))

    if not no_cleanup:
        lib.shutil.rmtree(source_dir)



"""Codecs imports."""

from .. import lib

__all__ = (
    'collections',
    'ipaddress',
    'numbers',
    'pathlib',
    'uuid',
    *lib.__all__
    )

import collections.abc
import ipaddress
import numbers
import pathlib
import uuid

from .. lib import *



"""Module objs unit tests."""

import pickle
import unittest

import fqr

from fqr . core import lib

from ... import mocking

from . import cfg


class Constants(cfg.Constants):
    """Constant values specific to unit tests in this file."""


class TestObjectBase(unittest.TestCase):
    """Fixture for testing `Object` base functionality."""

    def setUp(self) -> None:
        self.mcs = fqr.objects.metas.Meta
        self.cls = mocking.Derivative
        self.object_ = self.cls()
        self.trip = mocking.TripDeriv(
            str_field='123',
            )
        self.anti = mocking.AntiTripDeriv(
            str_field='321'
            )
        return super().setUp()

    def test_01_dict_functionality(self):
        """Test `Object.__getitem__()`."""

        self.assertTrue(
            self.object_['int_field']
            == self.object_.int_field
            == self.cls.int_field.default
            )

    def test_02_dict_functionality(self):
        """Test `Object.__getitem__()` raises KeyError if no key."""

        self.assertRaises(
            KeyError,
            lambda: self.object_['field_that_does_not_exist']
            )

    def test_03_iter(self):
        """Test `Object.__iter__()`."""

        self.assertEqual(dict(self.object_.__iter__()), dict(self.object_))

    def test_04_len(self):
        """Test `Object.__len__()`."""

        self.assertEqual(len(self.object_), len(self.cls.fields))

    def test_05_contains(self):
        """Test `Object.__contains__()`."""

        self.assertIn(self.object_.fields[0], self.object_)

    def test_06_ne(self):
        """Test `Object.__ne__()`."""

        self.assertFalse(self.object_ != self.object_)

    def test_07_lshift(self):
        """Test `Object.__lshift__()` correctly interpolates."""

        object_ = self.trip << self.anti
        self.assertNotEqual(object_.str_field, self.anti.str_field)

    def test_08_lshift(self):
        """Test `Object.__lshift__()` correctly interpolates."""

        object_ = self.trip << self.anti
        self.assertNotEqual(object_.other_field, self.trip.other_field)

    def test_09_rshift(self):
        """Test `Object.__rshift__()` correctly overwrites."""

        default = mocking.TripDeriv()
        object_ = self.trip >> default
        self.assertNotEqual(object_.str_field, default.str_field)

    def test_10_rshift(self):
        """Test `Object.__rshift__()` correctly overwrites."""

        object_ = self.trip >> self.anti
        self.assertEqual(object_.str_field, self.anti.str_field)

    def test_11_dict_functionality(self):
        """Test `Object.get()` returns default if no key."""

        self.assertIsNone(self.object_.get('field_that_does_not_exist'))

    def test_12_to_dict(self):
        """Test `Object.to_dict()`."""

        self.assertDictEqual(
            {
                fqr.core.strings.utl.snake_case_to_camel_case(k): self.trip[k]
                for k, v in self.trip.items()
                if type(v) in lib.t.get_args(fqr.core.typ.Primitive)
                },
            {
                k: v
                for k, v
                in self.trip.to_dict(camel_case=True).items()
                if type(v) in lib.t.get_args(fqr.core.typ.Primitive)
                }
            )

    def test_13_to_dict(self):
        """Test `Object.to_dict()`."""

        self.assertDictEqual(
            {k: v for k, v in self.object_ if v is not None},
            self.object_.to_dict(include_null=False)
            )

    def test_14_repr(self):
        """Test `Object.__repr__()`."""

        self.assertEqual(
            repr(self.trip),
            lib.json.dumps(
                dict(self.trip),
                default=fqr.core.strings.utl.convert_for_repr,
                indent=Constants.INDENT,
                sort_keys=True
                )
            )

    def test_15_invalid_obj_comparison_exc(self):
        """
        Test InvalidObjectComparisonError raised when comparing \
        different Objects.

        """

        self.assertRaises(
            fqr.objects.exc.InvalidObjectComparisonError,
            lambda: mocking.NewDeriv() << mocking.TripDeriv()
            )

    def tearDown(self) -> None:
        return super().tearDown()


class TestObject(unittest.TestCase):
    """Fixture for testing `Object`."""

    def setUp(self) -> None:
        self.cls = mocking.Derivative
        self.object_= self.cls(str_field='cba')
        return super().setUp()

    def test_01_dunder_get_cls(self):
        """Test fields.Field __get__."""

        self.assertIsInstance(self.cls.str_field, fqr.Field)

    def test_02_dunder_get_ins(self):
        """Test fields.Field __get__."""

        self.assertIsInstance(
            self.object_.str_field,
            self.cls.str_field.type_
            )

    def test_03_get_enumeration_from_fields(self):
        """Test we get enumerations from Object fields."""

        self.assertIsInstance(
            fqr.objects.utl.get_enumerations_from_fields(
                self.cls.__dataclass_fields__
                ),
            dict
            )

    def test_04_get_enumeration_from_fields(self):
        """Test we get enumerations from Object fields."""

        self.assertIn(None, self.object_.enumerations['from_dict_field'])

    def test_05_get_fields_for_hash(self):
        """Test we get hashable fields from Object fields."""

        self.assertIn('secondary_key', self.object_.hash_fields)

    def test_06_delitem(self):
        """Test `Object.__delitem__()`."""

        del self.object_['str_field']
        self.assertEqual(
            self.cls.str_field.default,
            self.object_.str_field
            )

    def test_07_delitem(self):
        """Test `Object.__delitem__()`."""

        self.assertRaises(
            KeyError,
            lambda: self.object_.__delitem__('field_that_does_not_exist')
            )

    def test_08_setitem(self):
        """Test `Object.__setitem__()`."""

        self.assertRaises(
            KeyError,
            lambda: self.object_.__setitem__(
                'field_that_does_not_exist',
                1
                )
            )

    def test_09_reversed(self):
        """Test `Object.__reversed__()`."""

        self.assertListEqual(
            sorted(self.object_.keys(), reverse=True),
            list(reversed(self.object_))
            )

    def test_10_serialization(self):
        """Test `Object.__getstate__()`."""

        pickled = pickle.dumps(self.object_)
        self.assertEqual(self.object_, pickle.loads(pickled))

    def test_11_ior(self):
        """Test `Object.__ior__()`."""

        new_obj = self.cls()
        new_obj |= self.object_

        other_new_obj = self.cls()
        other_new_obj.update(self.object_)

        self.assertEqual(new_obj, other_new_obj)

    def test_12_copy(self):
        """Test `Object.__ior__()`."""

        new_obj = self.object_.copy()

        self.assertIsNot(new_obj, self.object_)

    def test_13_fromkeys(self):
        """Test `Object.__fromkeys__()`."""

        self.assertEqual(self.cls(), self.cls.fromkeys(()))

    def test_14_pop(self):
        """Test `Object.pop()`."""

        self.assertIsNone(self.object_.pop('not_a_real_field', None))

    def test_15_pop(self):
        """Test `Object.pop()`."""

        self.assertRaises(
            KeyError,
            lambda: self.object_.pop('not_a_real_field')
            )


class TestObjectDocumentationExamples(unittest.TestCase):
    """Test examples provided in Object __doc__."""

    def setUp(self) -> None:
        self.instance_values = {
            'id': 'abc123',
            '_alternate_id': 'dog1',
            'name': 'Bob',
            'type': 'dog',
            'in': 'timeout',
            'is_tail_wagging': False
            }
        self.new_name = 'Buddy'
        self.cls = mocking.examples.Pet
        self.object_= self.cls(
            id='abc123',
            _alternate_id='dog1',
            name='Bob',
            type='dog',
            in_='timeout',
            is_tail_wagging=False
            )
        self.object_from_dict = self.cls(self.instance_values)
        self.object_from_camel_case = self.cls(
            {
                'id': 'abc123',
                'alternateId': 'dog1',
                'name': 'Bob',
                'type': 'dog',
                'in': 'timeout',
                'isTailWagging': False
                }
            )
        return super().setUp()

    def test_01_instantiation(self):
        """Test __init__."""

        self.assertEqual(self.object_, self.object_from_dict)

    def test_02_instantiation_with_case_conversion(self):
        """Test __init__."""

        self.assertEqual(self.object_from_camel_case, self.object_)

    def test_03_dict_functionality(self):
        """Test __getitem__."""

        key = 'type'
        self.assertEqual(self.object_[key], self.instance_values[key])

    def test_04_dict_functionality(self):
        """Test __setitem__."""

        key = 'name'
        self.object_[key] = self.new_name
        self.assertTupleEqual(
            (self.new_name, self.new_name),
            (self.object_.name, self.object_[key])
            )

    def test_05_dict_functionality(self):
        """Test setdefault."""

        key = 'name'
        object_= self.cls()
        object_.setdefault(key, self.new_name)
        object_.setdefault(key, self.instance_values[key])
        self.assertTrue(
            getattr(object_, key)
            == self.new_name
            != self.instance_values[key]
            )

    def test_06_dict_keys(self):
        """Test keys."""

        self.assertListEqual(
            sorted(self.object_.keys()),
            sorted(self.instance_values)
            )

    def test_07_dict_values(self):
        """Test values."""

        self.assertListEqual(
            list(self.object_.values()),
            list(self.instance_values.values())
            )

    def test_08_dict_items(self):
        """Test items."""

        self.assertListEqual(
            sorted(self.object_.items()),
            sorted(self.instance_values.items())
            )

    def test_09_dict_key_error(self):
        """Test cannot set undefined field."""

        self.assertRaises(
            KeyError,
            lambda: self.object_.setdefault(
                'field_that_does_not_exist',
                self.new_name
                )
            )



"""
Overview
========

**Author:** dan@1howardcapital.com

**Summary:** Zero-dependency python framework for object oriented development.
Implement _once_, document _once_, in _one_ place.

---

With fqr, you will quickly learn established best practice... \
or face the consequences of runtime errors that will break your code \
if you deviate from it.

Experienced python engineers will find a framework \
that expects and rewards intuitive magic method implementations, \
consistent type annotations, and robust docstrings.

Implement _pythonically_ with fqr and you will only ever need to: \
implement _once_, document _once_, in _one_ place.

---

Getting Started
---------------

### Installation

Install from command line, with pip:

`$ pip install fqr`

"""

__all__ = (
    'cli',
    'core',
    'docs',
    'log',
    'loggers',
    'objects',
    'Field',
    'Object'
    )

__version__ = '0.7.0'

from . import cli
from . import core
from . import docs
from . import loggers
from . import objects

from . loggers import log
from . objects import Field, Object



"""Docs constants."""

__all__ = (
    'Constants',
    )

from .. import core

from . import lib


class Constants(core.cfg.Constants):
    """Constant values specific to docs modules."""

    BASE_VERSION    = '0.0.0'
    DOCS_STATIC_DIR = lib.os.path.join(
        lib.os.path.split(lib.os.path.dirname(__file__))[0],
        'docs',
        'static',
        )

    try:
        DEFAULT_USER = lib.os.getlogin()
    except OSError:  # pragma: no cover
        DEFAULT_USER = '<UNSPECIFIED>'



"""Core exceptions."""

__all__ = (
    'BasePackageException',
    )

from . import lib
from . import typ


class BasePackageException(
    BaseException,
    lib.t.Generic[lib.Unpack[typ.ArgsType]]
    ):
    """
    Exception common to the entire package.

    ---

    Automatically handles serialization.

    """

    def __init__(self, msg: str, *args: lib.Unpack[typ.ArgsType]) -> None:
        """Instantiate `fqr` exception."""

        self._args = args
        super().__init__(msg)

    def __reduce__(
        self: typ.PackageExceptionType
        ) -> tuple[
            type[typ.PackageExceptionType],
            tuple[lib.Unpack[typ.ArgsType]]
            ]:
        return (
            self.__class__,
            self._args
            )



import json
import unittest

import fqr

from ... import mocking


class TestField(unittest.TestCase):
    """Fixture for testing the object."""

    def setUp(self) -> None:
        self.int_field = fqr.Field(
            name='int_field',
            default=1,
            type_=int,
            )
        self.str_field = fqr.Field(
            name='str_field',
            default='a',
            type_=str,
            )
        self.union_field = fqr.Field(
            name='union_field',
            default='a',
            type_=int | str,
            )
        self.blank_field = fqr.Field(
            name='str_field',
            type_=str,
            )
        return super().setUp()

    def test_01_hash(self):
        """Test __hash__."""

        self.assertTrue(
            (
                self.str_field.__field_hash__()
                == fqr.Field(dict(self.str_field)).__field_hash__()
                == hash(
                    ''.join(
                        (
                            fqr.Field.__name__,
                            repr(self.str_field.type_),
                            repr(self.str_field.default),
                            self.str_field.name
                            )
                        )
                    )
                ) and self.str_field != self.int_field
            )

    def test_02_validate_comparison_exc(self):
        """Test invalid comparison type error raised."""

        self.assertRaises(
            fqr.objects.exc.InvalidComparisonTypeError,
            lambda: self.str_field % 2
            )

    def test_03_validate_comparison_container_exc(self):
        """Test invalid container comparison error raised."""

        self.assertRaises(
            fqr.objects.exc.InvalidContainerComparisonTypeError,
            lambda: self.int_field % 2
            )

    def test_04_validate_comparison_union_exc(self):
        """Test invalid comparison type error raised for Unions."""

        self.assertRaises(
            fqr.objects.exc.InvalidComparisonTypeError,
            lambda: self.union_field != bytes()
            )

    def test_05_lshift_overwrite(self):
        """Test __lshift__ updates when passed type same as self."""

        interpolated = self.blank_field << self.str_field
        self.assertEqual(interpolated.default, self.str_field.default)

    def test_06_repr_overwrite(self):
        """Test __repr__ same as _fields.Field."""

        self.assertEqual(
            repr(mocking.Derivative.bool_field),
            repr(
                fqr.objects.fields.Field(dict(mocking.Derivative.bool_field))
                )
            )

    def test_07_field_set_type_error(self):
        """Test __setitem__ raises correct exc if Field has invalid type."""

        self.assertRaises(
            fqr.objects.exc.IncorrectTypeError,
            lambda: fqr.objects.metas.Meta.__setitem__(
                fqr.Field,
                'name',
                fqr.Field(name='name', type_=int)
                )
            )


class TestFieldValidation(unittest.TestCase):
    """Fixture for testing the object."""

    def setUp(self) -> None:
        self.cls = mocking.TripDeriv
        return super().setUp()

    def test_01_parse(self):
        """Test correct exc raised if not nullable."""

        self.assertRaises(
            fqr.objects.exc.TypeValidationError,
            lambda: self.cls.non_nullable_field.parse(None)
            )

    def test_02_parse(self):
        """Test None."""

        self.assertIsNone(self.cls.null_field.parse(None))

    def test_03_parse(self):
        """Test bool."""

        self.assertIs(self.cls.bool_field.parse('true'), True)

    def test_04_parse(self):
        """Test decimal."""

        self.assertIsInstance(
            self.cls.decimal_field.parse('4.5'),
            self.cls.decimal_field.type_
            )

    def test_05_parse(self):
        """Test str null."""

        self.assertIsNone(self.cls.null_field.parse('null'))

    def test_06_parse(self):
        """Test unknown str exc."""

        self.assertRaises(
            fqr.objects.exc.TypeValidationError,
            lambda: self.cls.null_field.parse('asdf')
            )

    def test_07_parse(self):
        """Test Object from dict."""

        self.assertIsInstance(
            self.cls.new_deriv.parse(dict(self.cls.new_deriv.default)),
            self.cls.new_deriv.type_
            )

    def test_08_parse(self):
        """Test Union."""

        self.assertIsInstance(
            self.cls.forward_ref_union_field.parse('39.8'),
            fqr.core.typ.utl.check.get_checkable_types(
                self.cls.forward_ref_union_field.type_
                )
            )

    def test_09_parse(self):
        """Test Union."""

        self.assertRaises(
            fqr.objects.exc.TypeValidationError,
            lambda: self.cls.forward_ref_union_field.parse(
                fqr.Field(type_=str)
                ),
            )

    def test_10_parse(self):
        """Test Object from json."""

        self.assertIsInstance(
            self.cls.new_deriv.parse(repr(self.cls.new_deriv.default)),
            self.cls.new_deriv.type_
            )

    def test_11_parse(self):
        """Test any dict."""

        self.assertEqual(
            self.cls.dict_field.parse(
                json.dumps(
                    self.cls.generic_dict_field.default
                    )
                ),
            self.cls.generic_dict_field.default
            )

    def test_12_parse(self):
        """Test generic dict."""

        self.assertRaises(
            fqr.objects.exc.TypeValidationError,
            lambda: self.cls.generic_dict_field.parse(
                json.dumps(self.cls.dict_field.default)
                ),
            )

    def test_13_parse(self):
        """Test any array."""

        self.assertEqual(
            self.cls.tuple_field.parse(
                json.dumps(
                    self.cls.generic_tuple_field.default
                    )
                ),
            self.cls.generic_tuple_field.default
            )

    def test_14_parse(self):
        """Test generic array."""

        self.assertRaises(
            fqr.objects.exc.TypeValidationError,
            lambda: self.cls.generic_tuple_field.parse(
                json.dumps(self.cls.tuple_field.default)
                ),
            )

    def test_15_set(self):
        """
        Test `Field__set__`.

        """

        obj = mocking.NewDeriv()

        def _fn():
            mocking.NewDeriv.generic_tuple_deriv_field.__set__(obj, 3)

        self.assertRaises(fqr.objects.exc.IncorrectTypeError, _fn)

    def test_16_set(self):
        """
        Test `Field__set__`.

        """

        obj = mocking.NewDeriv()

        def _fn():
            mocking.NewDeriv.generic_tuple_deriv_field.__set__(obj, '3')

        self.assertRaises(fqr.objects.exc.TypeValidationError, _fn)

    def test_17_set(self):
        """
        Test `Field__set__`.

        """

        obj = mocking.NewDeriv()
        mocking.NewDeriv.anti_field_1.__set__(obj, '3')

        self.assertEqual(obj.anti_field_1, '3')

    def test_18_set(self):
        """
        Test `Field__set__`.

        """

        obj = mocking.NewDeriv()
        mocking.NewDeriv.anti_field_2.__set__(obj, True)

        self.assertEqual(obj.anti_field_2, True)

    def test_19_parse(self):
        """
        Test parse returns `None` instead of error when set `False`.

        """

        self.assertIsNone(
            self.cls.generic_dict_field.parse(
                json.dumps(self.cls.dict_field.default),
                raise_validation_error=False
                )
            )

    def test_20_eq(self):
        """Test `Field.__eq__()`."""

        self.assertTrue(
            self.cls.generic_dict_field == self.cls.generic_dict_field
            )



"""Constant values specific to module unit tests."""

__all__ = (
    'Constants',
    )

from .. import cfg


class Constants(cfg.Constants):
    """Constant values specific to unit tests in this module."""

    INVALID_STRING_CASING_EXAMPLE = 'WRONG_cASasdfING9000man'



"""Metaclass module."""

__all__ = (
    'Meta',
    )

from ... import core

from .. import cfg
from .. import exc
from .. import lib
from .. import typ

from . import utl

if lib.t.TYPE_CHECKING:  # pragma: no cover
    from .. import fields as fields_


class Constants(cfg.Constants):
    """Constant values specific to this file."""


class Meta(type):
    """Base class constructor."""

    if lib.t.TYPE_CHECKING:  # pragma: no cover
        __annotations__: typ.SnakeDict
        __dict__: dict[typ.AnyString, lib.t.Any]
        __dataclass_fields__: lib.t.ClassVar[typ.DataClassFields]
        __heritage__: lib.t.ClassVar[tuple['Meta', ...]]

        enumerations: lib.t.ClassVar[dict[str, tuple[typ.Primitive, ...]]]
        fields: lib.t.ClassVar[typ.FieldsTuple]
        hash_fields: lib.t.ClassVar[typ.FieldsTuple]

    def __new__(
        mcs: type[typ.MetaType],
        __name: str,
        __bases: tuple[type, ...],
        __namespace: dict[typ.AnyString, lib.t.Any],
        **kwargs: lib.t.Any
        ) -> typ.MetaType:
        fields: typ.DataClassFields = {}
        heritage: tuple[type, ...] = __bases
        slots: list[typ.string[typ.snake_case]]
        _slots: tuple[typ.string[typ.snake_case], ...] | str = (
            __namespace.get(Constants.__SLOTS__, ())
            )
        if (
            isinstance(_slots, str)
            and core.strings.utl.is_snake_case_string(_slots)
            ):  # pragma: no cover
            slots = [_slots] if _slots != Constants.CLASS_AS_DICT else []
        elif core.strings.utl.is_snake_case_iterable(_slots):
            slots = [s for s in _slots if s != Constants.CLASS_AS_DICT]
        else:  # pragma: no cover
            slots = []
        module: str = __namespace.get(Constants.__MODULE__, '')
        annotations: typ.SnakeDict
        annotations = __namespace.pop(Constants.__ANNOTATIONS__, {})
        annotations |= {
            k: typ.utl.hint.resolve_type(
                v,
                lib.sys.modules[module].__dict__
                )
            for k, v
            in annotations.items()
            }

        base_count = 0
        for _base in reversed(__bases):
            if isinstance(_base, Meta):
                base_count += 1
                fields |= _base.__dataclass_fields__

        __namespace.pop(Constants.CLASS_AS_DICT, None)
        annotations.pop(Constants.CLASS_AS_DICT, None)

        if base_count > 1:
            from .. import objs
            common_annotations: typ.SnakeDict = {}
            common_base_names: list[str] = []
            common_bases: list[type[objs.Object]] = []
            common_namespace: dict[typ.AnyString, lib.t.Any] = {}
            common_slots: list[typ.string[typ.snake_case]] = []
            for _base in reversed(__bases):
                for __base in reversed(_base.__mro__):
                    if (
                        issubclass(__base, objs.Object)
                        and __base is not objs.Object
                        ):
                        if __base.__name__ not in common_base_names:
                            common_base_names.insert(0, __base.__name__)
                            common_bases.insert(0, __base)
                            for slot in __base.__slots__:
                                if slot not in common_slots:
                                    common_slots.append(slot)
                            common_annotations |= __base.__annotations__
                            common_namespace |= __base.__dict__
            common_namespace = {
                k: v
                for k, v
                in common_namespace.items()
                if (
                    utl.is_valid_keyword(k)
                    and k not in common_slots
                    )
                }
            common_namespace[Constants.__ANNOTATIONS__] = common_annotations
            common_namespace[Constants.__SLOTS__] = tuple(common_slots)
            common_base = Meta(
                Constants.DELIM_REBASE.join(common_base_names[:base_count]),
                (objs.Object, ),
                common_namespace
                )
            __bases = (common_base, )

        if module != Constants.OBJECTS_MODULE:
            base_fields = set(fields.keys())
            
            defaults, slots, fields = (
                utl.parse_new_namespace(
                    __namespace,
                    annotations,
                    module,
                    slots,
                    fields
                    )
                )

            for name in defaults:
                __namespace.pop(name)

            slots, fields = (
                utl.parse_new_annotations(
                    __namespace,
                    annotations,
                    module,
                    slots,
                    fields,
                    base_fields
                    )
                )

        fields_tuple = tuple(sorted(fields))

        namespace = {
            Constants.__SLOTS__: tuple(slots),
            **__namespace,
            }

        namespace[Constants.__ANNOTATIONS__] = annotations
        namespace[Constants.__DATACLASS_FIELDS__] = fields
        namespace[Constants.__HERITAGE__] = heritage

        namespace[Constants.FIELDS] = fields_tuple
        namespace[Constants.ENUMERATIONS] = (
            utl.get_enumerations_from_fields(fields)
            )
        if module != Constants.FIELDS_MODULE:
            namespace[Constants.HASH_FIELDS] = utl.get_fields_for_hash(fields)
        else:
            namespace[Constants.HASH_FIELDS] = ('name', )

        return super().__new__(
            mcs,
            __name,
            __bases,
            namespace,
            **kwargs
            )

    def __repr__(cls) -> str:
        """
        Return constructor represented as a neatly formatted JSON string.

        """

        return core.codecs.utl.serialize(cls)

    def __getattribute__(cls, __name: str) -> lib.t.Any:
        __fields: dict[str, 'typ.AnyField[lib.t.Any]'] = (
            type.__getattribute__(cls, '__dataclass_fields__')
            )
        if (field := __fields.get(__name)):
            return field
        elif __name == 'class_as_dict':  # pragma: no cover
            # This clause exists to address sphinx-doc error
            # where sphinx thinks this attribute is otherwise
            # more available / heritable than it really is.
            return {}
        else:
            return super().__getattribute__(__name)

    def __setattr__(
        cls,
        __name: str,
        __value: lib.t.Any
        ) -> lib.t.Optional[lib.Never]:
        if (
            (cname := core.strings.utl.cname_for(__name, cls.fields))
            and not typ.utl.check.is_field(__value)
            ):
            raise exc.IncorrectTypeError(__name, type(__value), __value)
        elif cname:
            cls[__name] = __value
            return None
        else:
            super().__setattr__(__name, __value)
            return None

    @lib.t.overload
    def __getitem__(
        cls,
        __key: typ.AnyString
        ) -> 'fields_.Field[lib.t.Any]' | lib.Never: ...
    @lib.t.overload
    def __getitem__(
        cls,
        __key: type[typ.AnyType]
        ) -> 'typ.Field[typ.AnyType]' | lib.Never: ...
    def __getitem__(
        cls,
        __key: typ.AnyType | type[typ.AnyType]
        ) -> lib.t.Union[
            'fields_.Field[lib.t.Any]',
            'typ.Field[typ.AnyType]',
            lib.Never
            ]:
        """Return value dict style."""

        if (
            isinstance(__key, str)
            and core.strings.utl.is_snake_case_string(__key)
            and (k := core.strings.utl.cname_for(__key, cls.fields))
            ):
            return cls.__dataclass_fields__[k]
        elif typ.utl.check.is_field_type(cls):
            return typ.Field(cls, __key)
        else:
            raise KeyError(__key)

    def __contains__(cls, __key: lib.t.Any) -> bool:
        """Return `True` if `__key` is a field for class."""

        return bool(core.strings.utl.cname_for(__key, cls.fields))

    def __iter__(cls) -> lib.t.Iterator[typ.string[typ.snake_case]]:
        """Iterate over field names."""

        for fname in cls.fields:
            yield fname

    def __instancecheck__(cls, __instance: lib.t.Any) -> bool:
        """Instance check that considers slotted heritage."""

        return (
            super().__instancecheck__(__instance)
            or cls in getattr(__instance, '__heritage__', ())
            )

    def __subclasscheck__(cls, __subclass: type[lib.t.Any]) -> bool:
        """Subclass check that considers slotted heritage."""

        return (
            super().__subclasscheck__(__subclass)
            or cls in getattr(__subclass, '__heritage__', ())
            )

    def __setitem__(
        cls,
        __key: str,
        __value: lib.t.Any
        ) -> lib.t.Optional[lib.Never]:
        """Set Field for key dict style."""

        if not typ.utl.check.is_field(__value):
            from .. import fields as fields_
            raise exc.IncorrectTypeError(__key, fields_.Field, __value)
        elif (
            (k := core.strings.utl.cname_for(__key, cls.fields))
            and __value['name'] != k
            ):
            raise exc.InvalidFieldRedefinitionError(__value['name'])
        elif k and not issubclass(
            __value['type'],
            typ.utl.check.get_checkable_types(
                (ftype := cls.__dataclass_fields__[k]['type'])
                )
            ):
            raise exc.IncorrectTypeError(k, ftype, __value['type'])
        elif k and not isinstance(
            __value['default'],
            typ.utl.check.get_checkable_types(
                (ftype := cls.__dataclass_fields__[k]['type'])
                )
            ):
            raise exc.IncorrectDefaultTypeError(
                k,
                ftype,
                __value['default']
                )
        elif k is not None:
            value_: 'fields_.Field[lib.t.Any]' = __value
            cls.__dataclass_fields__[k].update(value_)  # type: ignore[arg-type]
            return None
        else:
            raise exc.InvalidFieldAdditionError(__key)
